from deposit_gui.view.vusertools_elements.user_elements.user_element import (UserElement)

class ColumnBreak(UserElement):
	
	def __init__(self):
		
		UserElement.__init__(self)
	
	def to_markup(self):
		
		return "<ColumnBreak/>"



from deposit_gui.dgui.abstract_subview import AbstractSubview
from deposit_gui.view.vmdiarea_frames.query_frame import QueryFrame
from deposit_gui.view.vmdiarea_frames.external_frame import ExternalFrame
from deposit_gui.view.vmdiarea_frames.class_graph_frame import ClassGraphFrame

from PySide2 import (QtWidgets, QtCore, QtGui)

class VMDIArea(AbstractSubview, QtWidgets.QFrame):
	
	signal_window_activated = QtCore.Signal(bool)
	
	def __init__(self, on_drag_enter, on_drag_move, on_drop):
		
		AbstractSubview.__init__(self)
		QtWidgets.QFrame.__init__(self)
		
		self._outside_frames = []
		self._mdiarea = MDIArea(on_drag_enter, on_drag_move, on_drop)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self._mdiarea.subWindowActivated.connect(self.on_window_activated)
		
		self.layout().addWidget(self._mdiarea)
	
	def _add_frame(self, frame):
		# frame = AbstractMDIAreaFrame
		
		window = MdiSubWindow()
		window.setWidget(frame)
		window.setWindowTitle(frame.title())
		window.setWindowIcon(self.get_icon(frame.icon()))
		window.windowStateChanged.connect(lambda old_state, new_state: self.on_window_state_changed(old_state, new_state, window))
		self._mdiarea.addSubWindow(window)
		window.show()
	
	def _add_outside_frame(self, frame, *args, **kwargs):
		# frame = AbstractOutsideFrame
		
		frame.setWindowTitle(frame.title())
		frame.setWindowIcon(self.get_icon(frame.icon()))
		self._outside_frames.append(frame)
		frame.show()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def close_all(self):
		
		for frame in self._outside_frames:
			frame.close()
		self._outside_frames = []
		self._mdiarea.closeAllSubWindows()
	
	def set_background_text(self, text):
		
		self._mdiarea._background_text = text
		self.hide()
		self.show()
	
	
	def add_query_frame(self, query, cmodel, cview):
		
		query_frame = QueryFrame(query, cmodel, cview)
		self._add_frame(query_frame)
		
		return query_frame
	
	def get_query_frames(self):
		
		for window in self._mdiarea.subWindowList():
			frame = window.widget()
			if isinstance(frame, QueryFrame):
				yield frame
	
	
	def add_external_frame(self, source, *args, **kwargs):
		
		external_frame = ExternalFrame(source, *args, **kwargs)
		self._add_frame(external_frame)
		
		return external_frame
	
	def get_external_frames(self):
		
		for window in self._mdiarea.subWindowList():
			frame = window.widget()
			if isinstance(frame, ExternalFrame):
				yield frame
	
	
	def add_class_graph_frame(self, cmodel, cview):
		
		class_graph_frame = ClassGraphFrame(cmodel, cview)
		self._add_frame(class_graph_frame)
		
		return class_graph_frame
	
	def get_class_graph_frames(self):
		
		for window in self._mdiarea.subWindowList():
			frame = window.widget()
			if isinstance(frame, ClassGraphFrame):
				yield frame
	
	
	def deselect_all(self):
		
		for frame in self._outside_frames:
			frame.on_deactivate()
		for window in self._mdiarea.subWindowList():
			window.widget().on_deactivate()
	
	def get_current_frame(self):
		
		current = self._mdiarea.currentSubWindow()
		if current:
			current = current.widget()
			if current:
				return current
		return None
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot(object)
	def on_window_activated(self, window):
		
		self.signal_window_activated.emit(True)
	
	@QtCore.Slot(object, object, object)
	def on_window_state_changed(self, old_state, new_state, window):
		
		was_active = False
		is_active = False
		if old_state & QtCore.Qt.WindowActive:
			was_active = True
		if new_state & QtCore.Qt.WindowActive:
			is_active = True
		if was_active and not is_active:
			frame = window.widget()
			if frame is not None:
				frame.on_deactivate()
		if was_active != is_active:
			self.signal_window_activated.emit(is_active)

class MDIArea(QtWidgets.QMdiArea):
	
	def __init__(self, on_drag_enter, on_drag_move, on_drop):
		
		QtWidgets.QMdiArea.__init__(self)
		
		self._on_drag_enter = on_drag_enter
		self._on_drag_move = on_drag_move
		self._on_drop = on_drop
		self._background_text = ""
		
		self.setAcceptDrops(True)
		self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
		self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
	
	def paintEvent(self, event):

		QtWidgets.QMdiArea.paintEvent(self, event)

		if self._background_text:
			painter = QtGui.QPainter()
			td = QtGui.QTextDocument()
			
			painter.begin(self.viewport())
			painter.translate(QtCore.QPointF(30, 30))
			
			font = td.defaultFont()
			font.setPointSize(10)
			td.setDefaultFont(font)
			td.setHtml(self._background_text)
			td.drawContents(painter)
			
			painter.end()
	
	def dragEnterEvent(self, event):
		
		self._on_drag_enter(event)
	
	def dragMoveEvent(self, event):
		
		self._on_drag_move(event)
	
	def dropEvent(self, event):
		
		self._on_drop(event)

class MdiSubWindow(QtWidgets.QMdiSubWindow):

	def __init__(self):
		
		QtWidgets.QMdiSubWindow.__init__(self, flags = QtCore.Qt.SubWindow)
		
		self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
	
	def closeEvent(self, *args, **kwargs):
		
		frame = self.widget()
		if frame is not None:
			frame.on_close()
		
		QtWidgets.QMdiSubWindow.closeEvent(self, *args, **kwargs)



from deposit_gui.dgui.dconnect_tabs.dconnect_tab_db import DConnectTabDB
from deposit_gui.dgui.dconnect_tabs.dconnect_tab_db_rel import DConnectTabDBRel

from PySide2 import (QtWidgets, QtCore, QtGui)
from pathlib import Path
import os

class DSaveAsPostgresFrame(QtWidgets.QFrame):
	
	def __init__(self, dialog):
		
		QtWidgets.QFrame.__init__(self)
		
		self._dialog = dialog
		
		self.setMinimumWidth(600)
		self.setMinimumHeight(400)
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self.setStyleSheet('''
			QPushButton {margin: 3px 3px 3px 3px; padding: 5px 10px 5px 10px;}
		''')
		
		self._tab_db = DConnectTabDB(self)
		self._tab_dbrel = DConnectTabDBRel(self)
		
		self._tab_db.connect_button.hide()
		self._tab_db.delete_button.hide()
		self._tab_dbrel.connect_button.hide()
		self._tab_dbrel.delete_button.hide()
		
		self._tabs = QtWidgets.QTabWidget()
		self._tabs.addTab(self._tab_db, "PostgreSQL")
		self._tabs.addTab(self._tab_dbrel, "PostgreSQL Relational")
		
		self.layout().addWidget(self._tabs)
		
		self._dialog.set_title("Save As PostgreSQL")
		self._dialog.set_frame(self)
		self._dialog.setModal(True)
	
	def set_recent_connections(self, data):
		# data = [[url], [identifier, connstr], ...]
		
		self._tab_db.set_recent_connections(data)
		self._tab_dbrel.set_recent_connections(data)
	
	def on_connect(self, identifier = None, connstr = None, url = None, datasource = None):
		
		self._dialog.set_data(dict(
			identifier = identifier,
			connstr = connstr,
			url = url,
			datasource = datasource,
		))
		
		self._dialog.accept()
	
	def creating_enabled(self):
		
		return True
	
	def connect_caption(self):
		
		return "Connect"



from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.view.vquerytoolbar import VQueryToolbar

from PySide2 import (QtCore)

class CQueryToolbar(AbstractSubcontroller):
	
	def __init__(self, cmain, cview) -> None:
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._vquerytoolbar = VQueryToolbar(cview._view)
		
		self._vquerytoolbar.signal_entered.connect(self.on_query_entered)
	
	def set_query_text(self, text):
		
		self._vquerytoolbar.set_query_text(text)
	
	def get_query_text(self):
		
		return self._vquerytoolbar.get_query_text()
	
	@QtCore.Slot(str)
	def on_query_entered(self, querystr):
		
		self.cmain.cmdiarea.add_query(querystr)




from deposit_gui.view.vusertools_elements.dialog.dialog_frame import (DialogFrame)

from PySide2 import (QtWidgets, QtCore)

class DialogMultiGroup(QtWidgets.QGroupBox):
	
	entry_added = QtCore.Signal()
	entry_removed = QtCore.Signal(list) # [obj_id, ...]
	
	def __init__(self, cmodel, user_group):
		# user_group = MultiGroup
		
		self._cmodel = cmodel
		self.user_group = user_group
		self.controls_frame = None
		self._framesets = []
		
		QtWidgets.QGroupBox.__init__(self, self.user_group.label)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.controls_frame = QtWidgets.QFrame()
		self.controls_frame.setLayout(QtWidgets.QVBoxLayout())
		self.controls_frame.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().addWidget(self.controls_frame)
		
		button_frame = QtWidgets.QFrame()
		button_frame.setLayout(QtWidgets.QHBoxLayout())
		button_frame.layout().setContentsMargins(5, 5, 5, 5)
		button_frame.layout().addStretch()
		button_add = QtWidgets.QPushButton("Add Entry")
		button_add.clicked.connect(self.on_add_entry)
		button_frame.layout().addWidget(button_add)
		self.layout().addWidget(button_frame)
		
		self.add_entry()
		
		if self.user_group.stylesheet:
			self.setStyleSheet(self.user_group.stylesheet)
	
	def add_frameset(self):
		
		self._framesets.append(QtWidgets.QFrame())
		self._framesets[-1].setLayout(QtWidgets.QVBoxLayout())
		self.controls_frame.layout().addWidget(self._framesets[-1])
		return self._framesets[-1]
	
	def framesets(self):
		# returns [[DialogFrame, ...], ...]
		
		return [list(frameset.findChildren(DialogFrame, options = QtCore.Qt.FindDirectChildrenOnly)) for frameset in self._framesets]
	
	def add_frame(self, user_control):
		
		self._framesets[-1].layout().addWidget(DialogFrame(self._cmodel, user_control))
	
	def add_entry(self):
		
		if len(self._framesets) > 0:
			line = QtWidgets.QFrame()
			line.setFrameShape(QtWidgets.QFrame.HLine)
			line.setFrameShadow(QtWidgets.QFrame.Sunken)
			self.controls_frame.layout().addWidget(line)
		self.add_frameset()
		for member in self.user_group.members:
			self.add_frame(member)
		
		button_frame = QtWidgets.QWidget()
		button_frame.setLayout(QtWidgets.QHBoxLayout())
		button_frame.layout().setContentsMargins(0, 0, 0, 0)
		button_remove = QtWidgets.QPushButton("Remove")
		button_remove.clicked.connect(self.on_remove_entry)
		button_frame.layout().addStretch()
		button_frame.layout().addWidget(button_remove)
		self._framesets[-1].layout().addWidget(button_frame)
		
		return list(self._framesets[-1].findChildren(DialogFrame, options = QtCore.Qt.FindDirectChildrenOnly))
	
	def remove_entry(self, frameset):
		
		for idx, frm in enumerate(self._framesets):
			if frm == frameset:
				break
		
		obj_ids = set([])
		for frame in self._framesets[idx].findChildren(DialogFrame, options = QtCore.Qt.FindDirectChildrenOnly):
			obj_id = frame.ctrl.obj_id
			if obj_id is not None:
				obj_ids.add(obj_id)
			frame.set_value("", None)
		
		if idx > 0:
			
			# remove HLine after frameset
			idx_layout = self.controls_frame.layout().indexOf(frameset)
			line = self.controls_frame.layout().itemAt(idx_layout + 1)
			if line is not None:
				line = line.widget()
				self.controls_frame.layout().removeWidget(line)
				line.setParent(None)
			
			self.controls_frame.layout().removeWidget(frameset)
			frameset.setParent(None)
			del self._framesets[idx]
		
		return list(obj_ids)
	
	def clear(self):
		
		for frame in self.controls_frame.findChildren(QtWidgets.QFrame, options = QtCore.Qt.FindDirectChildrenOnly):
			if frame.__class__.__name__ != "QFrame":
				continue
			self.controls_frame.layout().removeWidget(frame)
			frame.setParent(None)
		self._framesets = []
		self.add_entry()
	
	@QtCore.Slot()
	def on_add_entry(self):
	
		self.add_entry()
		self.entry_added.emit()
	
	@QtCore.Slot()
	def on_remove_entry(self):
		
		obj_ids = self.remove_entry(self.sender().parent().parent())
		self.entry_removed.emit(obj_ids)
		



from deposit_gui.view.vusertools_elements.user_elements.user_element_list import (UserElementList)

from PySide2 import (QtCore)

class EntryForm(UserElementList):
	
	def __init__(self, label, vusertools):
		
		UserElementList.__init__(self, label, vusertools)
		
		self.setIcon(self._vusertools.get_icon("form.svg"))
		self.setToolTip("Entry Form: %s" % self.label)
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self._vusertools.open_entry_form(self)
	



from PySide2 import (QtWidgets, QtCore, QtGui)

class DialogAddRelation(QtWidgets.QFrame):
	
	def __init__(self, dialog, cmain, elements, label, is_class):
		
		QtWidgets.QFrame.__init__(self)
		
		self._dialog = dialog
		self._cmain = cmain
		self._elements = elements
		self._is_class = is_class
		self._targets = []
		
		self._dialog.set_title("Add %sRelation" % ("Class " if self._is_class else ""))
		self._dialog.setModal(False)
		self._dialog.set_button_box(True, True)
		self._dialog.set_enabled(False)
		
		self.setMinimumWidth(256)
		self.setLayout(QtWidgets.QVBoxLayout())
		labels = QtWidgets.QWidget()
		labels.setLayout(QtWidgets.QHBoxLayout())
		self.label_source = QtWidgets.QLabel("")
		self.label_relation = QtWidgets.QLineEdit(label)
		if not label:
			self.label_relation.setPlaceholderText("Relation name")
		self.label_target = QtWidgets.QLabel("")
		labels.layout().addWidget(self.label_source)
		labels.layout().addWidget(self.label_relation)
		labels.layout().addWidget(self.label_target)
		self.layout().addWidget(labels)
		
		self._cmain.signal_selection_changed.connect(self.on_selection_changed)
		
		self.update()
	
	def update(self):
		
		source_txt = " \u2192 "
		target_txt = " \u2192 "
		if self._is_class:
			source_txt = "%s \u2192 " % (", ".join([cls.name for cls in self._elements]))
			if self._targets:
				target_txt = " \u2192 %s" % (", ".join([cls.name for cls in self._targets]))
		else:
			source_txt = "obj(%s) \u2192 " % (", ".join([str(obj_id) for obj_id in sorted([obj.id for obj in self._elements])]))
			if self._targets:
				target_txt = " \u2192 obj(%s)" % (", ".join([str(obj_id) for obj_id in sorted([obj.id for obj in self._targets])]))
		self.label_source.setText(source_txt)
		self.label_target.setText(target_txt)
		
		label = self.get_label()
		
		self._dialog.set_enabled((len(self._targets) > 0) and (label != ""))
	
	def get_label(self):
		
		return self.label_relation.text().strip()
	
	def get_targets(self):
		
		return self._targets
	
	@QtCore.Slot()
	def on_selection_changed(self):
		
		if self._is_class:
			self._targets = list(self._cmain.get_selected_classes())
		else:
			self._targets = list(self._cmain.get_selected_objects())
		
		self.update()




from deposit_gui.view.vusertools_elements.user_elements.user_element import (UserElement)

class UserUnique(UserElement):
	
	def __init__(self, dclass):
		
		self.dclass = dclass
		
		UserElement.__init__(self)
	
	def to_dict(self):
		
		return dict(
			dclass = self.dclass,
			**UserElement.to_dict(self),
		)




'''
Generic class to be re-used also by Deposit in the future
'''

from PySide2 import (QtWidgets, QtCore, QtGui)

class DProgress(QtWidgets.QProgressDialog):
	
	def __init__(self, view):
		
		self._view = view
		
		QtWidgets.QProgressDialog.__init__(self, "", "Cancel", 0, 0, view, flags = QtCore.Qt.FramelessWindowHint)
		
		self.setWindowModality(QtCore.Qt.WindowModal)
		
		self.cancel()
		self.reset()
	
	def cancel_pressed(self):
		
		return self.wasCanceled()
	
	def show(self, text = ""):
		
		self.reset()
		
		self.setLabelText(text)
		
		QtWidgets.QProgressDialog.show(self)
		
		QtWidgets.QApplication.processEvents()
	
	def stop(self):
		
		self.cancel()
		self.reset()
	
	def update_state(self, text = None, value = None, maximum = None):
		
		if not self.isVisible():
			self.show()
		if text is not None:
			self.setLabelText(text)
		if value is not None:
			self.setValue(value)
		if maximum is not None:
			self.setMaximum(maximum)
		
		QtWidgets.QApplication.processEvents()




from deposit_gui.view.vusertools_elements.editor import editor_controls as EditorControls

from deposit.utils.fnc_serialize import (select_to_class_descr)

from PySide2 import (QtWidgets)

class EditorFrame(QtWidgets.QFrame):
	
	def __init__(self, element, form_editor, user_control = None):
		
		self.label_edit = None
		self.ctrl = None
		self.selected = False
		self.bold = False
		self.form_editor = form_editor
		
		QtWidgets.QFrame.__init__(self)
		
		self.setLayout(QtWidgets.QFormLayout())
		self.layout().setContentsMargins(10, 10, 10, 10)
		
		self.ctrl = getattr(EditorControls, element)(user_control)
		label = self.ctrl.user_control.label
		self.label_edit = QtWidgets.QLineEdit()
		if not label:
			self.label_edit.setPlaceholderText("Label")
		else:
			self.label_edit.setText(label)
		self.layout().addRow(self.label_edit, self.ctrl)
		
		self.bold = "font-weight: bold;" in self.ctrl.user_control.stylesheet
		
		self.update_stylesheet()
	
	def user_element(self):
		
		label = self.label_edit.text()
		select = select_to_class_descr(self.ctrl.select_text())
		
		if (not label) or (not select):
			return None
		
		self.ctrl.user_control.label = label
		self.ctrl.user_control.stylesheet = "QLabel {font-weight: bold;}" if self.bold else ""
		self.ctrl.user_control.dclass = select[0]
		self.ctrl.user_control.descriptor = select[1]
		return self.ctrl.user_control
	
	def update_stylesheet(self):
		
		stylesheet = "%s:hover {background: grey;}" % (self.__class__.__name__)
		if self.selected:
			stylesheet += " %s {border: 2px solid grey;} " % (self.__class__.__name__)
		self.setStyleSheet(stylesheet)
		if self.bold:
			self.label_edit.setStyleSheet("QLineEdit {font-weight: bold;}")
		else:
			self.label_edit.setStyleSheet("{}")
	
	def setSelected(self, state):
		
		self.selected = state
		self.update_stylesheet()
	
	def setBold(self, state):
		
		self.bold = state
		self.update_stylesheet()
	
	def mousePressEvent(self, event):
		
		state = not self.selected
		form_editor = self.form_editor
		if hasattr(form_editor, "form_editor"):
			form_editor = form_editor.form_editor
		form_editor.deselect_all()
		self.setSelected(state)
		form_editor.on_selection_changed()




from deposit_gui.view.vusertools_elements.editor.editor_form import (EditorForm)
from deposit_gui.view.vusertools_elements.editor.editor_query import (EditorQuery)
from deposit_gui.view.vusertools_elements.user_elements.user_tool import (UserTool)
from deposit_gui.view.vusertools_elements.user_elements.search_form import (SearchForm)
from deposit_gui.view.vusertools_elements.user_elements.entry_form import (EntryForm)
from deposit_gui.view.vusertools_elements.user_elements.query import (Query)

from PySide2 import (QtWidgets, QtCore)

class Manager(QtWidgets.QDialog):
	
	signal_export = QtCore.Signal(object)	# user_tool
	signal_import = QtCore.Signal()
	signal_del_tool = QtCore.Signal(list)	# [label, ...]
	
	def __init__(self, vusertools):
		
		QtWidgets.QDialog.__init__(self)
		
		self._vusertools = vusertools
		
		self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
		self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		self.setStyleSheet("QPushButton {Text-align:left;}")
		self.setWindowTitle("User Tool Manager")
		self.setMinimumWidth(300)
		self.setModal(True)
		self.setLayout(QtWidgets.QHBoxLayout())
		self.layout().setContentsMargins(10, 10, 10, 10)
		
		self.tool_list = QtWidgets.QListWidget()
		self.tool_list.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
		self.controls = QtWidgets.QFrame()
		self.controls.setLayout(QtWidgets.QVBoxLayout())
		self.controls.layout().setContentsMargins(0, 0, 0, 0)
		
		self.button_add_query = QtWidgets.QPushButton("Add Query")
		self.button_add_query.setIcon(self._vusertools.get_icon("add_query.svg"))
		self.button_add_query.setIconSize(QtCore.QSize(32, 32))
		self.button_add_query.clicked.connect(self.on_add_query)
		self.controls.layout().addWidget(self.button_add_query)
		
		self.button_add_search = QtWidgets.QPushButton("Add Search Form")
		self.button_add_search.setIcon(self._vusertools.get_icon("add_search.svg"))
		self.button_add_search.setIconSize(QtCore.QSize(32, 32))
		self.button_add_search.clicked.connect(self.on_add_search)
		self.controls.layout().addWidget(self.button_add_search)
		
		self.button_add_entry = QtWidgets.QPushButton("Add Entry Form")
		self.button_add_entry.setIcon(self._vusertools.get_icon("add_form.svg"))
		self.button_add_entry.setIconSize(QtCore.QSize(32, 32))
		self.button_add_entry.clicked.connect(self.on_add_entry)
		self.controls.layout().addWidget(self.button_add_entry)
		
		self.button_edit = QtWidgets.QPushButton("Edit")
		self.button_edit.setIcon(self._vusertools.get_icon("edit.svg"))
		self.button_edit.setIconSize(QtCore.QSize(32, 32))
		self.button_edit.clicked.connect(self.on_edit)
		self.controls.layout().addWidget(self.button_edit)
		
		self.button_delete = QtWidgets.QPushButton("Delete")
		self.button_delete.setIcon(self._vusertools.get_icon("delete.svg"))
		self.button_delete.setIconSize(QtCore.QSize(32, 32))
		self.button_delete.clicked.connect(self.on_delete)
		self.controls.layout().addWidget(self.button_delete)
		
		self.button_order_up = QtWidgets.QPushButton("Order Up")
		self.button_order_up.setIcon(self._vusertools.get_icon("up_small.svg"))
		self.button_order_up.setIconSize(QtCore.QSize(32, 32))
		self.button_order_up.clicked.connect(self.on_order_up)
		self.controls.layout().addWidget(self.button_order_up)
		
		self.button_order_down = QtWidgets.QPushButton("Order Down")
		self.button_order_down.setIcon(self._vusertools.get_icon("down_small.svg"))
		self.button_order_down.setIconSize(QtCore.QSize(32, 32))
		self.button_order_down.clicked.connect(self.on_order_down)
		self.controls.layout().addWidget(self.button_order_down)
		
		self.button_export = QtWidgets.QPushButton("Export")
		self.button_export.setIcon(self._vusertools.get_icon("export_deposit.svg"))
		self.button_export.setIconSize(QtCore.QSize(32, 32))
		self.button_export.clicked.connect(self.on_export)
		self.controls.layout().addWidget(self.button_export)
		
		self.button_import = QtWidgets.QPushButton("Import")
		self.button_import.setIcon(self._vusertools.get_icon("import.svg"))
		self.button_import.setIconSize(QtCore.QSize(32, 32))
		self.button_import.clicked.connect(self.on_import)
		self.controls.layout().addWidget(self.button_import)
		
		self.controls.layout().addStretch()
		
		self.layout().addWidget(self.tool_list)
		self.layout().addWidget(self.controls)
		
		self.populate()
		self.update()
		
		self.tool_list.itemSelectionChanged.connect(self.on_selection_changed)
		
	def start_query_editor(self, query_tool = None):
		
		dialog = EditorQuery(self._vusertools, query_tool)
		dialog.signal_add_tool.connect(lambda user_tool: self._vusertools.signal_add_tool.emit(user_tool))
		dialog.signal_update_tool.connect(lambda label, user_tool: self._vusertools.signal_update_tool.emit(label, user_tool))
		dialog.show()
	
	def start_form_editor(self, form_tool = None, entry = False):
		
		self._vusertools.form_editor = EditorForm(self._vusertools, form_tool, entry)
		self._vusertools.form_editor.signal_add_tool.connect(lambda user_tool: self._vusertools.signal_add_tool.emit(user_tool))
		self._vusertools.form_editor.signal_update_tool.connect(lambda label, user_tool: self._vusertools.signal_update_tool.emit(label, user_tool))
		self._vusertools.form_editor.show()
	
	def stop_form_editor(self):
		
		self._vusertools.form_editor.close()
		self._vusertools.form_editor = None
	
	def populate(self):
		
		self.tool_list.clear()
		for action in self._vusertools._toolbar.actions():
			if not issubclass(action.__class__, UserTool):
				continue
			item = QtWidgets.QListWidgetItem(action.icon(), action.text())
			item._tool = action
			self.tool_list.addItem(item)
	
	def get_selected(self):
		
		return [item._tool for item in self.tool_list.selectedItems()]
	
	def update(self):
		
		selected = self.get_selected()
		self.button_edit.setEnabled(len(selected) == 1)
		self.button_delete.setEnabled(len(selected) > 0)
		self.button_order_up.setEnabled(len(selected) == 1)
		self.button_order_down.setEnabled(len(selected) == 1)
		self.button_export.setEnabled(len(selected) == 1)
	
	@QtCore.Slot()
	def on_edit(self):
		
		tool = self.get_selected()[0]
		if isinstance(tool, Query):
			self.start_query_editor(tool)
		if isinstance(tool, SearchForm):
			self.start_form_editor(tool, entry = False)
		elif isinstance(tool, EntryForm):
			self.start_form_editor(tool, entry = True)
	
	@QtCore.Slot()
	def on_add_query(self):
		
		self.start_query_editor()
	
	@QtCore.Slot()
	def on_add_search(self):
		
		self.start_form_editor(entry = False)
	
	@QtCore.Slot()
	def on_add_entry(self):
		
		self.start_form_editor(entry = True)
	
	@QtCore.Slot()
	def on_delete(self):
		
		labels = [tool.label for tool in self.get_selected()]
		if not labels:
			return
		reply = QtWidgets.QMessageBox.question(self, "Delete Tool", "Delete %d tools?" % (len(labels)), QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
		if reply != QtWidgets.QMessageBox.Yes:
			return
		self.signal_del_tool.emit(labels)
	
	@QtCore.Slot()
	def on_order_up(self):
		
		pass
	
	@QtCore.Slot()
	def on_order_down(self):
		
		pass
	
	@QtCore.Slot()
	def on_export(self):
		
		self.signal_export.emit(self.get_selected()[0])
	
	@QtCore.Slot()
	def on_import(self):
		
		self.signal_import.emit()
	
	@QtCore.Slot()
	def on_selection_changed(self):
		
		self.update()
	
	@QtCore.Slot()
	def on_data_changed(self):
		
		self.update()
	
	@QtCore.Slot()
	def on_tools_changed(self):
		
		self.populate()



from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.view.vusertools import VUserTools

from deposit.utils.fnc_serialize import (load_user_tool, value_to_str)

from PySide2 import (QtCore)
from collections import defaultdict

class CUserTools(AbstractSubcontroller):
	
	def __init__(self, cmain, cview) -> None:
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._vusertools = VUserTools(cview._view, cmain.cmodel)
		self._vusertools.signal_search_submit.connect(self.on_search_submit)
		self._vusertools.signal_entry_submit.connect(self.on_entry_submit)
		self._vusertools.signal_entry_unlink.connect(self.on_entry_unlink)
		self._vusertools.signal_entry_remove.connect(self.on_entry_remove)
		self._vusertools.signal_import_tool.connect(self.on_import_tool)
		self._vusertools.signal_export_tool.connect(self.on_export_tool)
		self._vusertools.signal_add_tool.connect(self.on_add_tool)
		self._vusertools.signal_update_tool.connect(self.on_update_tool)
		self._vusertools.signal_del_tool.connect(self.on_del_tool)
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	
	@QtCore.Slot(str)
	def on_search_submit(self, querystr):
		
		self.cmain.cmdiarea.add_query(querystr)
	
	@QtCore.Slot(dict, dict, list, set, dict)
	def on_entry_submit(
			self, values, objects_existing, relations, unique, objects_loaded
		):
		#	values = {cls: {idx: {descr: value, ...}, ...}, ...}
		#	objects_existing = {cls: {idx: obj_id, ...}, ...}
		#	relations = [[cls1, rel, cls2], ...]
		#	unique = set(cls, ...)
		#	objects_loaded = {cls: [obj_id, ...], ...}
		
		# create / find objects
		objects = defaultdict(dict) # {cls: {idx: Object, ...}, ...}
		for cls in values:
			for idx in values[cls]:
				if not (True in [(value is not None) for value in values[cls][idx].values()]):
					continue
				found = False
				if cls not in unique:
					cls_ = self.cmain.cmodel.add_class(cls)
					for obj in cls_.get_members():
						found = True
						for descr in values[cls][idx]:
							if value_to_str(obj.get_descriptor(descr)) != values[cls][idx][descr]:
								found = False
								break
						if found:
							break
				if found:
					objects[cls][idx] = obj
				else:
					if idx in objects_existing[cls]:
						objects[cls][idx] = self.cmain.cmodel.get_object(objects_existing[cls][idx])
					else:
						objects[cls][idx] = self.cmain.cmodel.add_class(cls).add_member()
					for descr in values[cls][idx]:
						objects[cls][idx].set_descriptor(descr, values[cls][idx][descr])
				
				obj = objects[cls][idx]
				if (cls in objects_loaded) and (obj.id in objects_loaded[cls]):
					objects_loaded[cls].remove(obj.id)
		
		# create relations
		for cls1, rel, cls2 in relations:
			if not ((cls1 in objects) and (cls2 in objects)):
				continue
			for idx1 in objects[cls1]:
				obj1 = objects[cls1][idx1]
				for idx2 in objects[cls2]:
					obj2 = objects[cls2][idx2]
					if obj1 == obj2:
						continue
					obj1.add_relation(obj2, rel)
					if (cls1 in objects_loaded) and (obj1.id in objects_loaded[cls1]):
						objects_loaded[cls1].remove(obj1.id)
					if (cls2 in objects_loaded) and (obj2.id in objects_loaded[cls2]):
						objects_loaded[cls2].remove(obj2.id)
		
		# remove previous relations not present in the current form
		objects_form = set()
		for cls in objects:
			objects_form.update([obj.id for obj in objects[cls].values()])
		for cls in objects_loaded:
			objects_form.update(objects_loaded[cls])
		for cls1, rel, cls2 in relations:
			if cls1 in objects_loaded:
				for obj1 in objects_loaded[cls1]:
					obj1 = self.cmain.cmodel.get_object(obj1)
					for obj2, rel_ in list(obj1.get_relations()):
						if rel_ != rel:
							continue
						if obj2.id not in objects_form:
							continue
						if cls2 in obj2.get_class_names():
							obj1.del_relation(obj2, rel)
			if cls2 in objects_loaded:
				for obj2 in objects_loaded[cls2]:
					obj2 = self.cmain.cmodel.get_object(obj2)
					for obj1, rel_ in list(obj2.get_relations()):
						if self.cmain.cmodel.reverse_relation(rel_) != rel:
							continue
						if obj1.id not in objects_form:
							continue
						if cls1 in obj1.get_class_names():
							obj2.del_relation(obj1, rel_)
	
	@QtCore.Slot(dict, list, set, list)
	def on_entry_unlink(self, objects_existing, relations, unique, obj_ids):
		#	objects_existing = {cls: {idx: obj_id, ...}, ...}
		#	relations = [[cls1, rel, cls2], ...]
		#	unique = set(cls, ...)
		#	obj_ids = [obj_id, ...]; objects to be unlinked
		
		obj_lookup = defaultdict(set)  # {cls: {obj_id, ...}, ...}
		existing_objs = set([])
		for cls in objects_existing:
			vals = objects_existing[cls].values()
			obj_lookup[cls].update(vals)
			existing_objs.update(vals)
		
		to_unlink = set([])  # {(obj_id1, rel, obj_id2), ...}
		for obj_id in obj_ids:
			obj = self.cmain.cmodel.get_object(obj_id)
			obj_classes = set(obj.get_class_names())
			if unique.intersection(obj_classes):
				self.cmain.cmodel.del_object(obj_id)
			else:
				obj_relations = defaultdict(set)  # {rel: DObject, ...}
				for obj2, rel in obj.get_relations():
					obj_relations[rel].add(obj2)
				
				for cls1, rel, cls2 in relations:
					if (cls2 in obj_classes) and (("~" + rel) in obj_relations):
						rel = "~" + rel
						cls1, cls2 = cls2, cls1
					if (cls1 in obj_classes) and (rel in obj_relations):
						for obj2, rel_ in obj.get_relations():
							if rel_ != rel:
								continue
							if obj2 == obj:
								continue
							if obj2.id in existing_objs:
								if rel.startswith("~"):
									to_unlink.add((obj2, rel[1:], obj))
								else:
									to_unlink.add((obj, rel, obj2))
		for obj1, rel, obj2 in to_unlink:
			obj1.del_relation(obj2, rel)
	
	@QtCore.Slot(dict, set)
	def on_entry_remove(self, objects_existing, unique):
		#	objects_existing = {cls: {idx: obj_id, ...}, ...}
		#	unique = set(cls, ...)
		
		obj_ids = set([])
		for cls in objects_existing:
			obj_ids.update(objects_existing[cls].values())
		for obj_id in obj_ids:
			obj = self.cmain.cmodel.get_object(obj_id)
			if unique.intersection(obj.get_class_names()):
				self.cmain.cmodel.del_object(obj_id)
	
	@QtCore.Slot()
	def on_import_tool(self):
		
		path, format = self.cmain.cview.get_load_path("Import User Tool", "(*.txt)")
		if path is None:
			return
		data = load_user_tool(path)
		self.cmain.cmodel.add_user_tool(data)
	
	@QtCore.Slot(object)
	def on_export_tool(self, user_tool):
		
		path, format = self.cmain.cview.get_save_path("Export User Tool", "(*.txt)")
		if path is None:
			return
		with open(path, "w", encoding = "utf-8") as f:
			f.write(user_tool.to_markup())
	
	@QtCore.Slot(object)
	def on_add_tool(self, user_tool):
		
		self.cmain.cmodel.add_user_tool(user_tool.to_dict())
	
	@QtCore.Slot(str, object)
	def on_update_tool(self, label, user_tool):
		
		self.cmain.cmodel.del_user_tool(label)
		self.cmain.cmodel.add_user_tool(user_tool.to_dict())
	
	@QtCore.Slot(str)
	def on_del_tool(self, label):
		
		self.cmain.cmodel.del_user_tool(label)
	
	@QtCore.Slot()
	def on_data_changed(self):
		
		self._vusertools.on_data_changed()
	
	def on_close(self):
		
		self._vusertools.on_close()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	
	def populate_tools(self):
		
		self._vusertools.populate_tools(self.cmain.cmodel.get_user_tools())




from deposit_gui.view.vusertools_elements.user_elements.user_controls import (Select)

from deposit.utils.fnc_serialize import (select_to_class_descr)

from PySide2 import (QtWidgets)

class EditorSelect(QtWidgets.QFrame):
	
	def __init__(self, form_editor, user_select = None):
		
		self.label_edit = None
		self.user_select = None
		self.selected = False
		self.form_editor = form_editor
		
		QtWidgets.QFrame.__init__(self)
		
		if user_select is None:
			user_select = Select("", "", "", "")
		self.user_select = user_select
		
		self.setStyleSheet("%s:hover {background: grey;}" % (self.__class__.__name__))
		self.setLayout(QtWidgets.QHBoxLayout())
		self.layout().setContentsMargins(10, 10, 10, 10)
		
		dclass, descriptor = self.user_select.dclass, self.user_select.descriptor
		self.select = QtWidgets.QLineEdit()
		if (not dclass) or (not descriptor):
			self.select.setPlaceholderText("[Class].[Descriptor]")
		else:
			self.select.setText("[%s].[%s]" % (dclass, descriptor))
		self.layout().addWidget(self.select)
	
	def select_text(self):
		
		return self.select.text().strip()
	
	def user_element(self):
		
		select = select_to_class_descr(self.select_text())
		if not select:
			return None
		
		self.user_select.dclass = select[0]
		self.user_select.descriptor = select[1]
		return self.user_select
	
	def setSelected(self, state):
		
		if state:
			self.form_editor.deselect_all()
			self.setStyleSheet("%s {border: 2px solid grey;} %s:hover {background: grey;}" % (self.__class__.__name__, self.__class__.__name__))
		else:
			self.setStyleSheet("%s:hover {background: grey;}" % (self.__class__.__name__))
		self.selected = state
		self.form_editor.on_selection_changed()
	
	def mousePressEvent(self, event):
		
		self.setSelected(not self.selected)




from deposit_gui.controller.cmodel import CModel
from deposit_gui.controller.cview import CView
from deposit_gui.controller.cdialogs import CDialogs
from deposit_gui.controller.cnavigator import CNavigator
from deposit_gui.controller.cmdiarea import CMDIArea
from deposit_gui.controller.cactions import CActions
from deposit_gui.controller.cquerytoolbar import CQueryToolbar
from deposit_gui.controller.cusertools import CUserTools


from deposit import Store

from PySide2 import (QtWidgets, QtCore, QtGui)
import sys
import os

class Controller(QtCore.QObject):
	
	signal_selection_changed = QtCore.Signal()
	
	def __init__(self, store):
		
		QtCore.QObject.__init__(self)
		
		self._selected_objects = set()
		self._selected_classes = set()
		self._selected_relations = []
		self._selected_queryitems = set()
		
		self._is_subwindow = (store is not None)
		
		self.cmodel = CModel(self, store)
		self.cnavigator = CNavigator(self)
		self.cmdiarea = CMDIArea(self)
		self.cview = CView(self, self.cnavigator, self.cmdiarea)
		self.cdialogs = CDialogs(self, self.cview)
		self.cactions = CActions(self, self.cview)
		self.cquerytoolbar = CQueryToolbar(self, self.cview)
		self.cusertools = CUserTools(self, self.cview)
		
		self.cmodel.set_progress(self.cview.progress)
		self.cmodel.on_loaded()
		self.cview.log_message("Deposit GUI started")
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot()
	def on_navigator_activated(self):
		
		self.cmdiarea.deselect_all()
	
	@QtCore.Slot(bool)
	def on_mdiarea_activated(self, state):
		
		if state:
			self.cnavigator.deselect_all()
		self.cactions.update()
	
	@QtCore.Slot(list)
	def on_class_selected(self, classes):
		
		self._selected_queryitems = set()
		self._selected_objects = set()
		self._selected_classes = set(classes)
		self._selected_relations = []
		self.on_selection_changed()
	
	@QtCore.Slot(list)
	def on_object_selected(self, objects):
		
		self._selected_queryitems = set()
		self._selected_objects = set(objects)
		self._selected_classes = set()
		self._selected_relations = []
		self.on_selection_changed()
	
	@QtCore.Slot(list)
	def on_query_selected(self, items):
		
		self._selected_queryitems = set(items)
		self._selected_objects = set()
		self._selected_classes = set()
		self._selected_relations = []
		for item in self._selected_queryitems:
			if item.obj_id is not None:
				self._selected_objects.add(self.cmodel.get_object(item.obj_id))
		self.on_selection_changed()
	
	@QtCore.Slot(list)
	def on_relation_selected(self, relations):
		# relations = [(Source, Target, label), ...]
		
		self._selected_queryitems = set()
		self._selected_objects = set()
		self._selected_classes = set()
		self._selected_relations = relations
		self.on_selection_changed()
	
	def on_selection_changed(self):
		
		self.cactions.update()
		self.signal_selection_changed.emit()
	
	def on_close(self):
		
		if not self.check_save():
			return False
			
		self.cusertools.on_close()
		return True
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def get_selected_queryitems(self):
		# returns set(QueryItem, ...)
		
		return self._selected_queryitems
	
	def get_selected_objects(self):
		# returns set(DObject, ...)
		
		return self._selected_objects
	
	def get_selected_classes(self):
		# returns set(DClass, ...)
		
		return self._selected_classes
	
	def get_selected_relations(self):
		
		return self._selected_relations
	
	
	def check_save(self):
		
		if (not self._is_subwindow) and (not self.cmodel.is_saved()):
			reply = QtWidgets.QMessageBox.question(self.cview._view, 
				"Exit", "Save changes to database?",
				QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel
			)
			if reply == QtWidgets.QMessageBox.Yes:
				self.cactions.on_Save(True)
			elif reply == QtWidgets.QMessageBox.No:
				return True
			else:
				return False
		
		return True
	
	
	def open_in_external(self, path):
		
		if sys.platform in ["linux", "linux2", "darwin"]:
			return # TODO		
		if sys.platform.startswith("win"):
			if os.path.isfile(path):
				os.startfile(path)

	
	def close(self):
		
		self.cview.close()



version_info = (1, 4, 37)

__version__ = '.'.join(map(str, version_info))
__title__ = 'Deposit'
__date__ = "6.2.2023"

from deposit_gui.dgui_main import DGUIMain

from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.dgui.abstract_subview import AbstractSubview
from deposit_gui.dgui.dclickable_logo import DClickableLogo
from deposit_gui.dgui.dconnect_frame import DConnectFrame
from deposit_gui.dgui.dgraphics_view import DGraphicsView
from deposit_gui.dgui.dnotification import DNotification
from deposit_gui.dgui.dmain_window import DMainWindow
from deposit_gui.dgui.dgraph_view import DGraphView
from deposit_gui.dgui.dregistry import DRegistry
from deposit_gui.dgui.dcactions import DCActions
from deposit_gui.dgui.dcdialogs import DCDialogs
from deposit_gui.dgui.dcmodel import DCModel
from deposit_gui.dgui.dview import DView



from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.dgui.dmodel import DModel

from deposit import (DDateTime, DGeometry, DResource)
from deposit.datasource import AbstractDatasource

from deposit.utils.fnc_files import (as_url)
from deposit.utils.fnc_serialize import (try_numeric, value_to_str)
from deposit.query.parse import (remove_bracketed_selects, extract_expr_vars)

from PySide2 import (QtCore)
from collections import defaultdict
from natsort import natsorted

class DCModel(AbstractSubcontroller):
	
	def __init__(self, cmain, store = None):
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._progress = None
		
		self._model = DModel(store)
		
		self._model.signal_store_event.connect(self.on_store_event)
		self._model.signal_store_error.connect(self.on_error)
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot(int, list, list, object)
	def on_store_event(self, event, objects, classes, datasource):
		
		if event == DModel.EVENT_ADDED:
			self.on_added(objects, classes)
		elif event == DModel.EVENT_DELETED:
			self.on_deleted(objects, classes)
		elif event == DModel.EVENT_CHANGED:
			self.on_changed(objects, classes)
		elif event == DModel.EVENT_SAVED:
			self.on_saved(datasource)
		elif event == DModel.EVENT_LOADED:
			self.on_loaded()
		elif event == DModel.EVENT_LOCAL_FOLDER_CHANGED:
			self.on_local_folder_changed()
		elif event == DModel.EVENT_QUERIES_CHANGED:
			self.on_queries_changed()
		elif event == DModel.EVENT_USER_TOOLS_CHANGED:
			self.on_user_tools_changed()
		elif event == DModel.EVENT_SETTINGS_CHANGED:
			self.on_settings_changed()
	
	@QtCore.Slot(str)
	def on_error(self, message):
		
		pass
	
	def on_added(self, objects, classes):
		# elements = [DObject, DClass, ...]
		
		pass
	
	def on_deleted(self, objects, classes):
		# elements = [obj_id, name, ...]
		
		pass
	
	def on_changed(self, objects, classes):
		# elements = [DObject, DClass, ...]
		
		pass
	
	def on_saved(self, datasource):
		
		pass
	
	def on_loaded(self):
		
		pass
	
	def on_local_folder_changed(self):
		
		pass
	
	def on_queries_changed(self):
		
		pass
	
	def on_user_tools_changed(self):
		
		pass
	
	def on_settings_changed(self):
		
		pass
		
	
	# ---- get/set
	# ------------------------------------------------------------------------
	
	
	# ---- General
	# ------------------------------------------------------------------------
	def set_progress(self, progress):
		
		self._progress = progress
	
	def get_descriptor_names(self):
		
		return self._model.get_descriptor_names(ordered = True)
	
	def get_relation_labels(self):
		
		return self._model.get_relation_labels()
	
	def reverse_relation(self, label):
		
		if label.startswith("~"):
			return label[1:]
		return "~" + label
	
	def get_subgraph(self, objects):
		
		if self._progress is not None:
			self._progress.show("Extracting Subgraph")
		store = self._model.get_subgraph(
			objects, 
			progress = self._progress
		)
		if self._progress is not None:
			self._progress.stop()
		return store
	
	def get_thumbnail(self, resource, size = 256):
		
		return self._model.get_thumbnail(resource, size)
	
	def get_folder(self):
		
		return self._model.get_folder()
	
	def get_temp_folder(self, name):
		
		return self._model.get_temp_folder(name)
	
	def get_temp_copy(self, resource):
		
		return self._model.get_temp_copy(resource)
	
	def get_updated_url(self, resource):
		
		return self._model.get_updated_url(resource)
	
	def get_query(self, querystr, silent = False):
		
		if (not silent) and (self._progress is not None):
			self._progress.show("Processing Query")
		query = self._model.get_query(
			querystr, 
			progress = (None if silent else self._progress)
		)
		if (not silent) and (self._progress is not None):
			self._progress.stop()
		return query
	
	def has_local_folder(self):
		
		return self._model.has_local_folder()
	
	def set_local_folder(self, path):
		
		self._model.set_local_folder(path, progress = self._progress)
	
	def prune_resources(self):
		
		if self._progress is not None:
			self._progress.show("Pruning Resources")
		self._model.prune_resources()
		if self._progress is not None:
			self._progress.stop()
	
	def can_save(self):
		
		datasource = self._model.get_datasource()
		
		return datasource.__class__.__name__ != "Memory"
	
	def field_calculator(self, target, expr, rows):
		
		if (not target) or (not expr) or (not rows):
			return
		classes = set(self.get_class_names())
		descriptors = set(self.get_descriptor_names())
		expr, bracketed_selects = remove_bracketed_selects(
			expr, classes, descriptors
		)
		expr, vars = extract_expr_vars(
			expr.strip(), classes, descriptors, bracketed_selects
		)
		for obj_id in rows:
			obj = self.get_object(obj_id)
			values = dict(
				DDateTime = DDateTime,
				DGeometry = DGeometry,
				DResource = DResource,
			)
			for name in vars:
				values[name] = rows[obj_id].get(vars[name], None)
			value = eval(expr, values)
			if value is not None:
				obj.set_descriptor(target, value)
	
	def merge_objects(self, objects):
		# objects = [obj_id, ...]
		
		data = {}
		locations = {}
		relations = set()
		for obj_id in objects:
			obj = self.get_object(obj_id)
			for descr in obj.get_descriptors():
				if descr not in data:
					value = obj.get_descriptor(descr)
					if value is not None:
						data[descr] = value
						location = obj.get_location(descr)
						if location is not None:
							locations[descr] = location
		for obj_id in objects[1:]:
			obj = self.get_object(obj_id)
			for obj_tgt, label in obj.get_relations():
				weight = obj.get_relation_weight(obj_tgt, label)
				relations.add((obj_tgt, label, weight))
		obj = self.get_object(objects[0])
		for descr in data:
			obj.set_descriptor(descr, data[descr])
		for descr in locations:
			obj.set_location(descr, locations[descr])
		for obj_tgt, label, weight in relations:
			obj.add_relation(obj_tgt, label, weight)
		for obj_id in objects[1:]:
			self.del_object(obj_id)
	
	def merge_identical(self, objects):
		# objects = set(obj_id, ...)
		
		grouped = defaultdict(set)
		for obj_id in objects:
			obj = self.get_object(obj_id)
			key = [(descr.name, value_to_str(obj.get_descriptor(descr))) for \
				descr in obj.get_descriptors()]
			grouped[tuple(sorted(key, key = lambda item: item[0]))].add(obj)
		for key in grouped:
			if len(grouped[key]) > 1:
				obj = grouped[key].pop()
				relations = set()
				for obj2 in grouped[key]:
					for obj_tgt, label in obj.get_relations():
						weight = obj.get_relation_weight(obj_tgt, label)
						relations.add((obj_tgt, label, weight))
				for obj_tgt, label, weight in relations:
					obj.add_relation(obj_tgt, label, weight)
				for obj2 in grouped[key]:
					self.del_object(obj2)
	
	def duplicate(self, objects):
		# objects = set(obj_id, ...)
		
		for obj_id in objects:
			obj = self.get_object(obj_id)
			obj2 = self.add_object()
			for descr in obj.get_descriptors():
				obj2.set_descriptor(descr, obj.get_descriptor(descr))
				location = obj.get_location(descr)
				if location is not None:
					obj2.set_location(descr, location)
			for cls in obj.get_classes():
				cls.add_member(obj2)
	
	
	# ---- Object
	# ------------------------------------------------------------------------
	def add_object(self):
		
		return self._model.add_object()
	
	def get_object(self, obj_id):
		
		return self._model.get_object(obj_id)
	
	def del_object(self, obj_id):
		
		self._model.del_object(obj_id)
	
	def del_objects(self, obj_ids):
		
		self._model.blockSignals(True)
		if self._progress is not None:
			self._progress.show("Deleting")
			self._progress.update_state(value = 0, maximum = len(obj_ids))
		cnt = 1
		for obj_id in obj_ids:
			if self._progress is not None:
				if self._progress.cancel_pressed():
					break
			self._model.del_object(obj_id)
			if self._progress is not None:
				self._progress.update_state(value = cnt)
			cnt += 1
		self._model.blockSignals(False)
		self._model.on_deleted(obj_ids)
	
	
	# ---- Class
	# ------------------------------------------------------------------------
	def add_class(self, name):
		
		return self._model.add_class(name)
	
	def get_class(self, name):
		
		return self._model.get_class(name)
	
	def get_classes(self):
		
		return self._model.get_classes(ordered = True)
	
	def get_class_names(self, ordered = False):
		
		return self._model.get_class_names(ordered)
	
	def get_class_values(self, class_name, descriptor_name):
		
		values = set()
		cls = self._model.get_class(class_name)
		if cls is None:
			return []
		for obj in cls.get_members():
			val = obj.get_descriptor(descriptor_name)
			if val is not None:
				values.add(val)
		values = natsorted(values)
		
		return values
	
	def get_descriptor_names(self, ordered = False):
		
		return self._model.get_descriptor_names(ordered)
	
	def rename_class(self, cls, name):
		
		self._model.rename_class(cls, name)
	
	def rename_class_descriptor(self, descr, cls, name):
		
		self._model.rename_class_descriptor(descr, cls, name)
	
	def switch_order(self, cls1, cls2):
		
		self._model.switch_order(cls1, cls2)
	
	def del_class(self, name):
		
		self._model.del_class(name)
	
	def del_class_descriptor(self, descr, cls, class_only = False):
		
		self._model.del_class_descriptor(descr, cls, class_only)
	
	
	# ---- Import
	# ------------------------------------------------------------------------
	
	def add_data_row(self, 
		data: dict, 
		relations: set = set(), 
		unique: set = set(), 
		existing = {}, 
		return_added = False,
	):
		# add multiple objects with classes at once & automatically add relations 
		#	based on class relations or as specified in the relations attribute
		# data = {(Class name, Descriptor name): value, ...}
		# relations = {(Class name 1, label, Class name 2), ...}
		# unique = {Class name, ...}; always add a new object to classes 
		#	specified here, otherwise re-use objects with identical descriptors
		# existing = {Class name: Object, ...}
		#	use existing object for specified classes (i.e. just update descriptors)
		#
		# returns n_added or (n_added, added) if return_added == True
		#	added = {Class name: Object, ...}
		
		return self._model.add_data_row(
			data, relations, unique, existing, return_added
		)
	
	def import_data(self, get_data, n_rows, targets, relations, unique):
		# get_data(row, col) = value
		# targets = {col: (class_name, descriptor_name), ...}
		# relations = {(Class name 1, label, Class name 2), ...}
		# unique = {Class, ...}
		#
		# returns number of imported Objects
		
		n_added = 0
		for row in range(n_rows):
			data_row = {}  # {(Class name, Descriptor name): value, ...}
			for col in targets:
				value = try_numeric(get_data(row, col))
				if (value is None) or (value == ""):
					continue
				data_row[targets[col]] = value
			if data_row:
				n_added += self._model.add_data_row(data_row, relations, unique)
		return n_added
	
	def import_store(self, store, unique = set()):
		# unique = {Class name, ...}; always add a new object to classes 
		#	specified here, otherwise re-use objects with identical descriptors
		# progress = DProgress
		
		self._model.blockSignals(True)
		self._model.import_store(
			store, unique, progress = self._progress
		)
		self._model.blockSignals(False)
		self.on_changed(self._model.get_objects(), self._model.get_classes())
	
	
	# ---- User Tools
	# ------------------------------------------------------------------------
	def add_user_tool(self, user_tool):
		
		self._model.add_user_tool(user_tool)
	
	def get_user_tools(self):
		
		return self._model.get_user_tools()
	
	def del_user_tool(self, label):
		
		self._model.del_user_tool(label)
	
	
	# ---- Saved Query
	# ------------------------------------------------------------------------
	def add_saved_query(self, title, querystr):
		
		self._model.add_saved_query(title, querystr)
	
	def get_saved_query(self, title):
		
		return self._model.get_saved_query(title)
	
	def get_saved_queries(self):
		
		return self._model.get_saved_queries()
	
	def del_saved_query(self, title):
		
		self._model.del_saved_query(title)
	
	
	# ---- Persistence
	# ------------------------------------------------------------------------
	def has_auto_backup(self):
		
		return self._model.has_auto_backup()
	
	def set_auto_backup(self, state):
		
		self._model.set_auto_backup(state)
	
	def get_datasource(self):
		
		return self._model.get_datasource()
	
	def get_datasource_name(self):
		
		return self._model.get_datasource().get_name()
	
	def save(self, *args, **kwargs):
		
		if self._progress is not None:
			self._progress.show("Saving")
		if self._model.save(progress = self._progress, *args, **kwargs):
			if self._progress is not None:
				self._progress.stop()
			self.update_recent(kwargs)
			return True
		if self._progress is not None:
			self._progress.stop()
		return False
	
	def load(self, *args, **kwargs):
		# datasource = Datasource or format
		
		if self._progress is not None:
			self._progress.show("Loading")
		if self._model.load(progress = self._progress, *args, **kwargs):
			if self._progress is not None:
				self._progress.stop()
			self.update_recent(kwargs)
			return True
		if self._progress is not None:
			self._progress.stop()
		return False
	
	def is_saved(self):
		
		return self._model.is_saved()




from deposit_gui.dgui.abstract_subview import AbstractSubview
from deposit_gui.dgui.dmain_window import DMainWindow
from deposit_gui.dgui.dstatus_bar import DStatusBar
from deposit_gui.dgui.dprogress import DProgress
from deposit_gui.dgui.dlogging import DLogging

from deposit.utils.fnc_serialize import (
	encrypt_connstr,
	decrypt_connstr,
)

from PySide2 import (QtWidgets, QtCore, QtGui)
import traceback
import json
import sys
import os

class DView(AbstractSubview, DMainWindow):
	
	APP_NAME = "DView"
	REG_PREFIX = ""
	VERSION = "0.0.0"
	
	def __init__(self):
		
		self._current_geometry = None
		self._previous_geometry = None
		
		AbstractSubview.__init__(self)
		DMainWindow.__init__(self)
		
		self.progress = DProgress(self)
		self.statusbar = DStatusBar(self)
		self.logging = DLogging(self.APP_NAME, self.VERSION)
		
		self.set_title()
		self.setStatusBar(self.statusbar)
		
		sys.excepthook = self.exception_event
	
	
	# public methods
	
	def set_title(self, name: str = None) -> None:
		
		title = self.APP_NAME
		if name is None:
			self.setWindowTitle(title)
		else:
			self.setWindowTitle("%s - %s" % (name, title))
	
	def set_icon(self, name):
		
		self.setWindowIcon(self.get_icon(name))
	
	def get_recent_dir(self):
		
		return self.registry.get(self.REG_PREFIX + "recent_dir")
	
	def set_recent_dir(self, path):
		
		self.registry.set(self.REG_PREFIX + "recent_dir", path)
	
	def get_recent_connections(self):
		
		rows = self.registry.get(self.REG_PREFIX + "recent")
		if rows:
			rows = json.loads(rows)
			collect = []
			for row in rows:
				if len(row) == 2:
					identifier, connstr = row
					collect.append([identifier, decrypt_connstr(connstr)])
				else:
					collect.append(row)
			return collect
		return []
	
	def add_recent_connection(self, url = None, identifier = None, connstr = None):
		
		data = self.get_recent_connections()
		row = None
		if url is not None:
			row = [url]
		elif (identifier is not None) and (connstr is not None):
			row = [identifier, encrypt_connstr(connstr)]
		if row is None:
			return
		if row in data:
			data.remove(row)
		data = [row] + data
		self.registry.set(self.REG_PREFIX + "recent", json.dumps(data))
	
	def clear_recent_connections(self):
		
		self.registry.set(self.REG_PREFIX + "recent", "")
	
	def get_active_window(self):
		
		active = self
		for window in QtWidgets.QApplication.topLevelWidgets():
			if not isinstance(window, QtWidgets.QMainWindow):
				continue
			if window.isActiveWindow():
				active = window
				break
		return active
	
	
	def show_information(self, caption, text):
		
		QtWidgets.QMessageBox.information(
			self.get_active_window(), caption, text
		)
	
	def show_warning(self, caption, text):
		
		QtWidgets.QMessageBox.warning(self.get_active_window(), caption, text)
	
	def show_question(self, caption, text):
		
		reply = QtWidgets.QMessageBox.question(
			self.get_active_window(), caption, text
		)
		
		return reply == QtWidgets.QMessageBox.Yes
	
	def show_input_dialog(self, caption, text, value = "", **kwargs):
		
		text, ok = QtWidgets.QInputDialog.getText(
			self.get_active_window(), caption, text, text = value, **kwargs
		)
		if text and ok:
			return text
		return None
	
	def show_item_dialog(self, caption, text, items, editable = False):
		
		name, ok = QtWidgets.QInputDialog.getItem(
			self.get_active_window(), caption, text, items, editable
		)
		if name and ok:
			return name
		return None
	
	
	# events
	
	def exception_event(self, typ, value, tb):
		
		text = "Exception: %s: %s\nTraceback: %s" % (
			str(typ), str(value)[:512], "".join(traceback.format_tb(tb))
		)
		self.logging.append(text)
		print(text)
	
	
	# overriden QMainWindow methods
	
	def keyPressEvent(self, event):
		
		for action in self.findChildren(QtWidgets.QAction):
			if action.isEnabled() and (
				action.shortcut() == QtGui.QKeySequence(
					event.key()+int(event.modifiers())
				)
			):
				action.trigger()
		
		QtWidgets.QMainWindow.keyPressEvent(self, event)
	
	def closeEvent(self, event):
		
		DMainWindow.closeEvent(self, event)




from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.view.vmdiarea import VMDIArea
from deposit_gui.view.vmdiarea_frames.query_frame_elements.query_item import QueryItem
from deposit.store.abstract_delement import AbstractDElement
from deposit.utils.fnc_serialize import (try_numeric)
from deposit.utils.fnc_files import (url_to_path)
from deposit import DResource
from deposit import externalsource as Externalsource
from deposit import Store

from PySide2 import (QtWidgets, QtCore)
import copy
import json
import os

class CMDIArea(AbstractSubcontroller):
	
	def __init__(self, cmain) -> None:
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._vmdiarea = VMDIArea(self.on_drag_enter, self.on_drag_move, self.on_drop)
		
		self._vmdiarea.signal_window_activated.connect(self.cmain.on_mdiarea_activated)
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	
	@QtCore.Slot(object)
	def on_query_add_object(self, query):
		
		cls = query.main_class
		if cls == "*":
			return
		if cls is None:
			self.cmain.cmodel.add_object()
		else:
			cls = self.cmain.cmodel.get_class(cls)
			if cls:
				cls.add_member()
	
	@QtCore.Slot()
	def on_query_del_object(self):
		
		objects = self.cmain.get_selected_objects()
		if objects:
			self.cmain.cdialogs.open("DelObjects", objects)
	
	@QtCore.Slot()
	def on_query_del_descriptor(self):
		
		for item in self.cmain.get_selected_queryitems():
			if (item.obj_id is None) or (item.class_name is None) or (item.descriptor_name is None):
				continue
			obj = self.cmain.cmodel.get_object(item.obj_id)
			obj.del_descriptor(item.descriptor_name)
	
	@QtCore.Slot(object)
	def on_query_activated(self, item):
		
		if isinstance(item.value, DResource):
			value0 = copy.copy(item.value)
			url = self.cmain.cmodel.get_updated_url(item.value)
			if url is None:
				self.cmain.cview.show_warning(
					"File not found",
					"Could not locate file:\n%s" % (value0.url),
				)
				item.value = value0
				return
			self.cmain.open_in_external(url_to_path(url))
	
	@QtCore.Slot(object, object)
	def on_query_edited(self, item, value):
		
		descr = self.cmain.cmodel.get_class(item.descriptor_name)
		obj = self.cmain.cmodel.get_object(item.obj_id)
		if (descr is None) or (obj is None):
			return
		# TODO try json.dumps(value) for dicts and lists
		# TODO detect datetime in ISO format
		if not isinstance(value, AbstractDElement):
			value = try_numeric(value)
			if value == "":
				value = None
		obj.set_descriptor(descr, value)
	
	@QtCore.Slot(object, str)
	def on_query_drop_url(self, item, url):
		
		descr = self.cmain.cmodel.get_class(item.descriptor_name)
		obj = self.cmain.cmodel.get_object(item.obj_id)
		if (descr is None) or (obj is None):
			return
		obj.set_resource_descriptor(descr, url)
	
	@QtCore.Slot(object)
	def on_class_activated(self, cls):
		
		self.add_query("SELECT [%s].*" % (cls.name))
	
	@QtCore.Slot(str)
	def on_class_link(self, name):
		
		self.add_query("SELECT [%s].*" % (name))
	
	@QtCore.Slot(int, str, str)
	def on_relation_link(self, obj_id, label, name_tgt):
		
		self.add_query("SELECT [%s].*, OBJ(%d).* RELATED OBJ(%d).[%s].[%s]" % (name_tgt, obj_id, obj_id, label, name_tgt))
	
	@QtCore.Slot(int, str, str)
	def on_relation_unlink(self, obj_id, label, name_tgt):
		
		obj = self.cmain.cmodel.get_object(obj_id)
		to_del = []
		for obj_tgt, label_ in obj.get_relations():
			if (label_ == label) and obj_tgt.has_class(name_tgt):
				to_del.append(obj_tgt)
		for obj_tgt in to_del:
			obj.del_relation(obj_tgt, label)
	
	def get_drag_data(self, event):
		# returns type, data; type = "query_items" / "url" / None
		
		mimedata = event.mimeData()
		if "application/deposit" in mimedata.formats():
			rows = mimedata.data("application/deposit")
			rows = rows.data().decode("utf-8").strip()
			if not rows:
				cb = QtWidgets.QApplication.clipboard()
				mimedata = cb.mimeData()
				if "application/deposit" in mimedata.formats():
					rows = mimedata.data("application/deposit")
					rows = rows.data().decode("utf-8").strip()
			if rows:
				items = []
				rows = json.loads(rows)
				for row in rows:
					items.append(QueryItem(None).from_dict(row))
				return "query_items", items
		
		if mimedata.hasUrls():
			for url in mimedata.urls():
				url = str(url.toString())
				if url.split(".")[-1].lower() in ["json", "pickle", "xlsx", "csv", "shp"]:
					return "url", url
		
		return None, None
	
	def on_drag_enter(self, event):
		
		typ, data = self.get_drag_data(event)
		if typ is None:
			event.ignore()
			return
		event.accept()
	
	def on_drag_move(self, event):
		
		typ, data = self.get_drag_data(event)
		if typ is None:
			event.ignore()
			return
		event.accept()
	
	def on_drop(self, event):
		
		typ, data = self.get_drag_data(event)
		if typ == "url":
			url = "%s" % (data)
			ext = url.split(".")[-1].lower()
			if ext in ["json", "pickle"]:
				self.cmain.cdialogs.open("ConfirmLoad", url)
				return
			self.add_external(url)
			return
		
		if typ == "query_items":
			items = []
			datasource = None
			for item in data:
				if item.datasource.get("_id", None) == id(self.cmain.cmodel):
					return
				if item.datasource.get("datasource", "Memory") == "Memory":
					continue
				datasource = item.datasource
				items.append(item)
			if datasource is None:
				self.cmain.cview.show_warning("Import Error", "Cannot import from an unsaved database.")
			else:
				self.cmain.cview.progress.show("Loading")
				store = Store(keep_temp = True)
				store.load(progress = self.cmain.cview.progress, **datasource)
				self.cmain.cview.progress.stop()
				objects = []
				for item in items:
					objects.append(store.get_object(item.obj_id_row))
				sub_store = store.get_subgraph(objects)
				if not store.get_objects():
					return
				self.cmain.cdialogs.open("ImportStore", sub_store)
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	
	def get_current_frame(self):
		
		return self._vmdiarea.get_current_frame()
	
	def update_queries(self, objects, classes):
		
		for query_frame in self._vmdiarea.get_query_frames():
			query_frame.update_query(objects, classes)
	
	def update_class_graphs(self):
		
		for class_graph_frame in self._vmdiarea.get_class_graph_frames():
			class_graph_frame.update_classes()
	
	def set_background_text(self, text):
		
		self._vmdiarea.set_background_text(text)
	
	def add_query(self, querystr):
		
		query = self.cmain.cmodel.get_query(querystr)
		query_frame = self._vmdiarea.add_query_frame(query, self.cmain.cmodel, self.cmain.cview)
		query_frame.signal_query_selected.connect(self.cmain.on_query_selected)
		query_frame.signal_query_activated.connect(self.on_query_activated)
		query_frame.signal_object_selected.connect(self.cmain.on_object_selected)
		query_frame.signal_relation_selected.connect(self.cmain.on_relation_selected)
		query_frame.signal_class_link.connect(self.on_class_link)
		query_frame.signal_relation_link.connect(self.on_relation_link)
		query_frame.signal_relation_unlink.connect(self.on_relation_unlink)
		query_frame.signal_add_object.connect(self.on_query_add_object)
		query_frame.signal_del_object.connect(self.on_query_del_object)
		query_frame.signal_del_descriptor.connect(self.on_query_del_descriptor)
		query_frame.signal_edited.connect(self.on_query_edited)
		query_frame.signal_drop_url.connect(self.on_query_drop_url)
		self.cmain.cquerytoolbar.set_query_text(querystr)
	
	def add_class_graph(self):
		
		class_graph_frame = self._vmdiarea.add_class_graph_frame(self.cmain.cmodel, self.cmain.cview)
		class_graph_frame.signal_class_selected.connect(self.cmain.on_class_selected)
		class_graph_frame.signal_relation_selected.connect(self.cmain.on_relation_selected)
		class_graph_frame.signal_activated.connect(self.on_class_activated)
	
	def add_external(self, url):
		
		source = os.path.splitext(url)[-1].strip(".").upper()
		if not hasattr(Externalsource, source):
			return
		self._vmdiarea.add_external_frame(source, url = url)
	
	def deselect_all(self):
		
		self._vmdiarea.deselect_all()
	
	def close_all(self):
		
		self._vmdiarea.close_all()




class AbstractQueryTab(object):
	
	def update_query(self, *args, **kwargs):
		
		pass
	
	def apply_filter(self, *args, **kwargs):
		
		pass
	
	def get_row_count(self):
		
		return 0
	
	def on_close(self):
		
		pass
	
	def on_selected(self):
		
		pass




from PySide2 import (QtWidgets, QtCore, QtGui)

class DialogColumn(QtWidgets.QFrame):
	
	def __init__(self):
		
		QtWidgets.QFrame.__init__(self)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().addStretch()
		
#		self.setStyleSheet("DialogColumn {border: 1px solid gray;}")
	
	def add_widget(self, widget):
		
		self.layout().insertWidget(self.layout().count() - 1, widget)



from deposit_gui.view.vusertools_elements.user_elements.user_tool import (UserTool)

class UserElementList(UserTool):
	
	def __init__(self, label, vusertools):
		
		self.elements = []
		UserTool.__init__(self, label, vusertools)
	
	def to_markup(self):
		
		return UserTool.to_markup(self) + "\n".join([element.to_markup() for element in self.elements])
	
	def to_dict(self):
		
		return dict(
			elements = [element.to_dict() for element in self.elements],
			**UserTool.to_dict(self),
		)



from deposit_gui.view.vmdiarea_frames.query_frame_elements.abstract_query_tab import AbstractQueryTab
from deposit_gui.dgui.dgraph_view import (DGraphView, Node, NodeWithAttributes, NodeWithSimpleAttributes)
from deposit.utils.fnc_files import (sanitize_filename)

from PySide2 import (QtWidgets, QtCore, QtGui)
from pathlib import Path
import os

class QueryTabGraphLazy(AbstractQueryTab, QtWidgets.QWidget):
	
	def __init__(self, queryframe):
		
		QtWidgets.QWidget.__init__(self)
		
		self._queryframe = queryframe
		self._query = queryframe._query
	
	def get_row_count(self):
		
		return len(self._query)
	
	def deselect_all(self):
		
		pass
	
class QueryTabGraph(AbstractQueryTab, QtWidgets.QMainWindow):
	
	def __init__(self, queryframe, objects):
		
		QtWidgets.QMainWindow.__init__(self)
		
		self._queryframe = queryframe
		self._query = queryframe._query
		
		self._actions = {} # {name: QAction, ...}
		self._positions = {}  # {node_id: (x, y), ...}
		self._show_attributes = 1  # 0 = show nodes only; 1 = show descriptor values; 2 = show descriptor names and values
		self._objects = objects
		
		self._graph_view = DGraphView()
		self._graph_view.signal_node_activated.connect(self.on_activated)
		self._graph_view.signal_selected.connect(self.on_graph_selected)
		
		central_widget = QtWidgets.QWidget(self)
		central_widget.setLayout(QtWidgets.QVBoxLayout())
		central_widget.layout().setContentsMargins(0, 0, 0, 0)
		self.setCentralWidget(central_widget)
		central_widget.layout().addWidget(self._graph_view)
		
		self.toolbar = self.addToolBar("Graph")
		actions = [
			["node_view", "Show Nodes", "nodes.svg"],
			["value_view", "Show Values", "attributes_simple.svg"],
			["descriptor_view", "Show Descriptors with Values", "attributes.svg"],
			["reset_layout", "Re-arrange Layout", "geometry.svg"],
			["#separator", None, None],
			["zoom_in", "Zoom In", "zoom_in.svg"],
			["zoom_out", "Zoom Out", "zoom_out.svg"],
			["zoom_reset", "Zoom Reset", "zoom_reset.svg"],
			["#separator", None, None],
			["save_pdf", "Save As PDF", "save_pdf.svg"],
		]
		
		for name, text, icon in actions:
			if name == "#separator":
				self.toolbar.addSeparator()
			else:
				self._actions[name] = QtWidgets.QAction(self._queryframe.get_icon(icon), text, self)
				self._actions[name].setData(name)
				self.toolbar.addAction(self._actions[name])
		
		self._actions["node_view"].setCheckable(True)
		self._actions["value_view"].setCheckable(True)
		self._actions["descriptor_view"].setCheckable(True)
		self._actions["node_view"].setChecked(False)
		self._actions["value_view"].setChecked(True)
		self._actions["descriptor_view"].setChecked(False)
		
		self.toolbar.actionTriggered.connect(self.on_tool_triggered)
		
		self.populate_graph()
	
	def update_query(self):
		
		self.populate_graph()
	
	def populate_graph(self):
		
		self._graph_view.clear()
		
		if self._objects is None:
			return
		
		self._queryframe._cview.progress.show("Drawing Tree")
		
		nodes = []  # [AbstractNode, ...]
		edges = []  # [[source_id, target_id, label], ...]
		for obj in self._objects:
			
			descriptors = []
			if self._show_attributes > 0:
				for descr in obj.get_descriptors():
					value = obj.get_descriptor(descr)
					if value.__class__.__name__ == "DDateTime":
						value = value.isoformat
					elif value.__class__.__name__ == "DGeometry":
						value = value.geometry_type
					elif value.__class__.__name__ == "DResource":
						value = value.filename
					else:
						value = str(value)
					descriptors.append([descr.name, value])
			
			if self._show_attributes == 0:
				nodes.append(Node(obj.id, str(obj.id)))
			elif self._show_attributes == 1:
				nodes.append(NodeWithSimpleAttributes(obj.id, str(obj.id), descriptors))
			elif self._show_attributes == 2:
				nodes.append(NodeWithAttributes(obj.id, str(obj.id), descriptors))
			
			for obj_tgt, label in obj.get_relations():
				if label.startswith("~"):
					continue
				if obj_tgt in self._objects:
					edges.append([obj.id, obj_tgt.id, label])
		
		self._graph_view.populate(nodes = nodes, edges = edges, positions = self._positions, progress = self._queryframe._cview.progress)
		
		self._queryframe._cview.progress.stop()
	
	def apply_filter(self, objects):
		
		self._objects = objects
		if self.isVisible():
			self.populate_graph()
	
	def show(self):
		
		QtWidgets.QMainWindow.show(self)
		
		self.populate_graph()
	
	def get_row_count(self):
		
		return len(self._query)
	
	def clearSelection(self):
		
		self.deselect_all()
	
	def selectAll(self):
		
		self._graph_view.select_all()
	
	def deselect_all(self):
		
		self._graph_view.deselect_all()
	
	@QtCore.Slot(object)
	def on_tool_triggered(self, action):
		
		fnc_name = "on_%s" % str(action.data())
		if hasattr(self, fnc_name):
			getattr(self, fnc_name)()
	
	@QtCore.Slot(object)
	def on_activated(self, obj_id):
		
		self._queryframe.on_object_activated(obj_id)
	
	@QtCore.Slot()
	def on_graph_selected(self):
		
		self.on_selected()
	
	def on_selected(self):
		
		nodes, edges = self._graph_view.get_selected()
		self._queryframe.on_object_selected(nodes)
		if nodes:
			return
		self._queryframe.on_relation_selected(edges)
		
	
	# Toolbar actions
	
	def on_node_view(self):
		
		self.update_view(0)
	
	def on_value_view(self):
		
		self.update_view(1)
	
	def on_descriptor_view(self):
		
		self.update_view(2)
	
	def update_view(self, show_attributes):
		
		self._show_attributes = show_attributes
		for n, name in [(0, "node_view"), (1, "value_view"), (2, "descriptor_view")]:
			self._actions[name].blockSignals(True)
			self._actions[name].setChecked(show_attributes == n)
			self._actions[name].blockSignals(False)
		self.populate_graph()
	
	def on_reset_layout(self):
		
		self._positions = {}
		self.populate_graph()
	
	def on_zoom_in(self):
		
		self._graph_view.scale(1.1, 1.1)
	
	def on_zoom_out(self):
		
		self._graph_view.scale(0.9, 0.9)
	
	def on_zoom_reset(self):
		
		self._graph_view.reset_zoom()
	
	def on_save_pdf(self):
		
		name = None
		if self._query.main_class is not None:
			name = self._query.main_class.strip(".")
		filename = sanitize_filename(name + ".pdf", default = "objects.pdf")
		
		recent_dir = self._queryframe._cview.get_recent_dir()
		if not recent_dir:
			recent_dir = str(Path.home())
		path = os.path.join(recent_dir, filename)
		path, _ = QtWidgets.QFileDialog.getSaveFileName(None, "Save As Adobe PDF", path, "Adobe PDF (*.pdf)")
		if not path:
			return
		self._graph_view.save_pdf(path)
		self._queryframe._cview.set_recent_dir(os.path.dirname(path))




from PySide2 import (QtWidgets, QtCore, QtGui, QtSvg)
from svgelements import (
	SVG, Group, Image, Path, Polygon, Polyline, Line
)
import base64
import re

def svg_element_to_coords(element, scale = 1):
	
	coords = []
	if isinstance(element, Path):
		for segment in element:
			if isinstance(segment, Line):
				if not coords:
					coords.append([
						segment.start.x * scale,
						segment.start.y * scale,
					])
				coords.append([
					segment.end.x * scale, 
					segment.end.y * scale,
				])
		if coords:
			if coords[-1] == coords[0]:
				coords = coords[:-1]
	
	elif isinstance(element, Polygon):
		coords = [[point.x * scale, point.y * scale] for \
			point in element.points]
	
	elif isinstance(element, Polyline):
		coords = [[point.x * scale, point.y * scale] for \
			point in element.points]
	
	return coords

def svg_to_raster(path_svg, path_raster, 
	scale = 1,
	dpi = 72,
	width = None,
	height = None,
	max_side = None,
	stroke_width = None,
):
	
	def _coords_to_path(coords):
		
		path = QtGui.QPainterPath()
		path.moveTo(coords[0][0], coords[0][1])
		for point in coords[1:]:
			path.lineTo(point[0], point[1])
		return path	
		
	with open(path_svg, "r", encoding="utf-8") as f:
		src_svg = f.read()
	
	collect = {}  # {lap_id: [i, j, coords, transform, x0, y0], ...}
	
	for m_img in re.finditer("<image (.*?)/>", src_svg):
		i_img, j_img = m_img.start(0), m_img.end(0)
		lap_id = None
		for m_id in re.finditer(''' id="(.*?)"''', src_svg[i_img:j_img]):
			i_id, j_id = m_id.start(0) + 5, m_id.end(0) - 1
			lap_id = src_svg[i_img + i_id : i_img + j_id]
			break
		if lap_id is None:
			continue
		for m_data in re.finditer(''' xlink:href="(.*?)"''', src_svg[i_img:j_img]):
			collect[lap_id] = [
				i_img + m_data.start(0) + 13,
				i_img + m_data.end(0) - 1,
				None,
				None,
				None,
				None,
			]
	
	svg = SVG.parse(path_svg)
	for element in svg.elements():
		if isinstance(element, SVG):
			continue
		
		if isinstance(element, Group):
			if element.id not in collect:
				continue
			clip_path = [el for el in element.clip_path if \
				(el.__class__ in [Path, Polygon, Polyline])]
			if clip_path:
				collect[element.id][2] = svg_element_to_coords(clip_path[0], 1)
		
		elif isinstance(element, Image):
			if element.id not in collect:
				continue
			transform = element.transform
			x0, y0 = element.x, element.y
			collect[element.id][3:] = [transform, x0, y0]
	
	idxs = []
	for lap_id in collect:
		if len(collect[lap_id]) != 6:
			continue
		idxs.append(collect[lap_id])
	idxs = sorted(idxs, key = lambda row: row[:2])[::-1]
	
	for i, j, coords, transform, x0, y0 in idxs:
		
		data = src_svg[i:j]
		
		path = None
		if not data.startswith("data:image/"):
			continue
		
		image = QtGui.QImage()
		if path is not None:
			image.load(path)
		else:			
			image.loadFromData(base64.b64decode(data[23:]), "JPG")
		if image.isNull():
			continue
		
		transform = QtGui.QTransform(
			transform.a,
			transform.b,
			transform.c,
			transform.d,
			transform.e,
			transform.f,
		)
		
		pixmap = QtGui.QPixmap(image)
		
		collect = []
		for x, y in coords:
			point = transform.inverted()[0].map(QtCore.QPointF(x, y))
			collect.append([point.x() - x0, point.y() - y0])
		coords = collect
		
		w = pixmap.width()
		h = pixmap.height()
		path = _coords_to_path(coords)
		outline_path = QtWidgets.QGraphicsPathItem(path)
		outline_path.setBrush(QtCore.Qt.white)
		mask = QtGui.QPixmap(w, h)
		mask.fill(QtCore.Qt.black)
		painter = QtGui.QPainter(mask)
		outline_path.paint(painter, QtWidgets.QStyleOptionGraphicsItem())
		painter.end()
		mask = mask.createMaskFromColor(QtCore.Qt.white, QtCore.Qt.MaskOutColor)
		pixmap.setMask(mask)
		image = pixmap.toImage()
		ba = QtCore.QByteArray()
		buffer = QtCore.QBuffer(ba)
		buffer.open(QtCore.QIODevice.WriteOnly)
		image.save(buffer, "PNG")
		data = "data:image/png;base64," + ba.toBase64().data().decode("ascii")
		
		src_svg = src_svg[:i] + data + src_svg[j:]
	
	if stroke_width is not None:
		stroke_width = str(stroke_width)
		collect = ""
		j_last = 0
		for m in re.finditer(''' stroke-width="(.*?)"''', src_svg):
			i, j = m.start(0) + 15, m.end(0)
			collect += src_svg[j_last:i] + stroke_width + "\""
			j_last = j
		if j_last < len(src_svg):
			collect += src_svg[j_last:]
		src_svg = collect
	
	renderer = QtSvg.QSvgRenderer()
	renderer.load(QtCore.QXmlStreamReader(src_svg))
	
	res_mul = dpi / 25.4 # to dots/mm
	size = renderer.defaultSize()
	
	if (width is None) and (height is None) and (max_side is None):
		w = size.width() * scale * res_mul * 50/177
		h = size.height() * scale * res_mul * 50/177
	else:
		w, h = size.width(), size.height()
		
		if max_side is not None:
			width, height = None, None
			if w > h:
				width = max_side
			else:
				height = max_side
		
		if (width is not None) and (height is not None):
			w, h = width, height
		elif width is not None:
			h = width * (h / w)
			w = width
		else:
			w = height * (w / h)
			h = height
	
	image = QtGui.QImage(QtCore.QSize(w, h), QtGui.QImage.Format_RGB32)
	image.setDotsPerMeterX(dpi * 39.37)
	image.setDotsPerMeterY(dpi * 39.37)
	image.fill(QtGui.QColor(QtCore.Qt.white).rgb())
	painter = QtGui.QPainter(image)
	renderer.render(painter)
	painter.end()
	image.save(path_raster)




from deposit_gui.view.vnavigator_elements.class_list import ClassList

from PySide2 import (QtWidgets, QtCore, QtGui)

class ClassWidget(QtWidgets.QWidget):
	
	def __init__(self, vnavigator):
		
		QtWidgets.QWidget.__init__(self, vnavigator)
		
		self._vnavigator = vnavigator
		self._classlist = ClassList(self._vnavigator)
		self._footer = ClassFooter(self._vnavigator)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self.layout().addWidget(self._classlist)
		
		self.layout().addWidget(self._footer)
		
		self.update_footer()
	
	def populate(self, classes):
		
		self._classlist.populate(classes)
	
	def get_selected(self):
		
		return self._classlist.get_selected()
	
	def get_selected_is_descriptor(self):
		
		return self._classlist.get_selected_is_descriptor()
	
	def get_selected_parent(self):
		
		return self._classlist.get_selected_parent()
	
	def get_items_around_selected(self):
		
		return self._classlist.get_items_around_selected()
	
	def select_one_above(self):
		
		self._classlist.select_one_above()
	
	def select_one_below(self):
		
		self._classlist.select_one_below()
	
	def deselect_all(self):
		
		self._classlist.clearSelection()
	
	def update_footer(self):
		
		selected = self.get_selected()
		enabled1 = (len(selected) == 1) and (selected[0] != "!*")
		is_descriptor = self.get_selected_is_descriptor()
		enabled2 = enabled1 and not is_descriptor
		
		self._footer.button_add_desciptor.setEnabled(enabled2)
		self._footer.button_del_desciptor.setEnabled(is_descriptor)
		self._footer.button_rename.setEnabled(enabled1)
		self._footer.button_up.setEnabled(enabled1)
		self._footer.button_down.setEnabled(enabled1)
		self._footer.button_remove.setEnabled((len([name for name in selected if name != "!*"]) > 0))
	
class ClassFooter(QtWidgets.QFrame):
	
	def __init__(self, vnavigator):

		QtWidgets.QFrame.__init__(self)
		
		self._vnavigator = vnavigator
		
		self.setLayout(QtWidgets.QHBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().setSpacing(6)
		
		self.button_up = QtWidgets.QToolButton(self)
		self.button_up.setIcon(self._vnavigator.get_icon("up_small.svg"))
		self.button_up.setIconSize(QtCore.QSize(32, 32))
		self.button_up.setToolTip("Order Class Up")
		self.button_up.clicked.connect(self._vnavigator.on_class_up)
		self.layout().addWidget(self.button_up)
		
		self.button_down = QtWidgets.QToolButton(self)
		self.button_down.setIcon(self._vnavigator.get_icon("down_small.svg"))
		self.button_down.setIconSize(QtCore.QSize(32, 32))
		self.button_down.setToolTip("Order Class Down")
		self.button_down.clicked.connect(self._vnavigator.on_class_down)
		self.layout().addWidget(self.button_down)
		
		self.layout().addStretch()
		
		self.button_add = QtWidgets.QToolButton(self)
		self.button_add.setIcon(self._vnavigator.get_icon("add_class.svg"))
		self.button_add.setIconSize(QtCore.QSize(32, 32))
		self.button_add.setToolTip("Add Class")
		self.button_add.clicked.connect(self._vnavigator.on_class_add)
		self.layout().addWidget(self.button_add)
		
		self.button_add_desciptor = QtWidgets.QToolButton(self)
		self.button_add_desciptor.setIcon(self._vnavigator.get_icon("add_descriptor.svg"))
		self.button_add_desciptor.setIconSize(QtCore.QSize(32, 32))
		self.button_add_desciptor.setToolTip("Add Descriptor")
		self.button_add_desciptor.clicked.connect(self._vnavigator.on_class_add_descriptor)
		self.layout().addWidget(self.button_add_desciptor)
		
		self.button_del_desciptor = QtWidgets.QToolButton(self)
		self.button_del_desciptor.setIcon(self._vnavigator.get_icon("remove_descriptor.svg"))
		self.button_del_desciptor.setIconSize(QtCore.QSize(32, 32))
		self.button_del_desciptor.setToolTip("Remove Descriptor")
		self.button_del_desciptor.clicked.connect(self._vnavigator.on_class_del_descriptor)
		self.layout().addWidget(self.button_del_desciptor)
		
		self.button_rename = QtWidgets.QToolButton(self)
		self.button_rename.setIcon(self._vnavigator.get_icon("edit_class.svg"))
		self.button_rename.setIconSize(QtCore.QSize(32, 32))
		self.button_rename.setToolTip("Rename Class")
		self.button_rename.clicked.connect(self._vnavigator.on_class_rename)
		self.layout().addWidget(self.button_rename)
		
		self.button_remove = QtWidgets.QToolButton(self)
		self.button_remove.setIcon(self._vnavigator.get_icon("remove_class.svg"))
		self.button_remove.setIconSize(QtCore.QSize(32, 32))
		self.button_remove.setToolTip("Remove Class")
		self.button_remove.clicked.connect(self._vnavigator.on_class_remove)
		self.layout().addWidget(self.button_remove)



from deposit_gui.dgui.dview import DView

from deposit_gui import __version__, __title__

from PySide2 import (QtWidgets, QtCore, QtGui)

class View(DView):
	
	APP_NAME = __title__
	VERSION = __version__
	
	def __init__(self, vnavigator, vmdiarea) -> None:
		
		DView.__init__(self)
		
		self._close_callback = None
		
		central_widget = QtWidgets.QWidget(self)
		central_widget.setLayout(QtWidgets.QVBoxLayout())
		central_widget.layout().setContentsMargins(0, 0, 0, 0)
		self.setCentralWidget(central_widget)
		
		self._tool_window = QtWidgets.QMainWindow()
		self._tool_window.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
		central_widget.layout().addWidget(self._tool_window)
		
		splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
		splitter.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		central_widget.layout().addWidget(splitter)
		
		left_frame = QtWidgets.QFrame()
		left_frame.setLayout(QtWidgets.QVBoxLayout())
		left_frame.layout().setContentsMargins(0, 0, 0, 0)
		
		splitter.addWidget(vnavigator)
		splitter.addWidget(vmdiarea)
		
		self.setWindowIcon(self.get_icon("dep_cube.svg"))
		
	def menuBar(self):
		# HACK to stop MDI Subwindows move controls to MenuBar when maximized
		
		return self._tool_window.menuBar()
	
	def addToolBar(self, title):
		# HACK to stop MDI Subwindows move controls to MenuBar when maximized
		
		return self._tool_window.addToolBar(title)
	
	def addToolBarBreak(self, *args, **kwargs):
		# HACK to stop MDI Subwindows move controls to MenuBar when maximized
		
		self._tool_window.addToolBarBreak(*args, **kwargs)
	
	def closeEvent(self, event):
		
		if self._close_callback is not None:
			if not self._close_callback():
				event.ignore()
				return
		DView.closeEvent(self, event)



from deposit_gui.view.vusertools_elements.user_elements.user_controls import (Unique)

from PySide2 import (QtWidgets)

class EditorUnique(QtWidgets.QFrame):
	
	def __init__(self, form_editor, user_unique = None):
		
		self.label_edit = None
		self.user_unique = None
		self.selected = False
		self.form_editor = form_editor
		
		QtWidgets.QFrame.__init__(self)
		
		if user_unique is None:
			user_unique = Unique("", "", "")
		self.user_unique = user_unique

		self.setStyleSheet("%s:hover {background: grey;}" % (self.__class__.__name__))
		self.setLayout(QtWidgets.QHBoxLayout())
		self.layout().setContentsMargins(10, 10, 10, 10)
		
		self.unique = QtWidgets.QLineEdit()
		if not self.user_unique.dclass:
			self.unique.setPlaceholderText("Class")
		else:
			self.unique.setText(self.user_unique.dclass)
		self.layout().addWidget(self.unique)
	
	def select_text(self):
		
		return self.unique.text().strip()
	
	def user_element(self):
		
		unique = self.select_text()
		if not unique:
			return None
		
		self.user_unique.dclass = unique
		return self.user_unique
	
	def setSelected(self, state):
		
		if state:
			self.form_editor.deselect_all()
			self.setStyleSheet("%s {border: 2px solid grey;} %s:hover {background: grey;}" % (self.__class__.__name__, self.__class__.__name__))
		else:
			self.setStyleSheet("%s:hover {background: grey;}" % (self.__class__.__name__))
		self.selected = state
		self.form_editor.on_selection_changed()
	
	def mousePressEvent(self, event):
		
		self.setSelected(not self.selected)




from deposit_gui.view.vusertools_elements.dialog import dialog_controls as DialogControls

from PySide2 import (QtWidgets)

class DialogFrame(QtWidgets.QFrame):
	
	def __init__(self, cmodel, user_control):
		# user_control = UserElement
		
		self._cmodel = cmodel
		self.user_control = user_control
		self.ctrl = None
		
		QtWidgets.QFrame.__init__(self)
		
		self.ctrl = getattr(DialogControls, user_control.__class__.__name__)(self._cmodel, self.user_control)
		
		self.setLayout(QtWidgets.QFormLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().setRowWrapPolicy(QtWidgets.QFormLayout.WrapLongRows)
		label = QtWidgets.QLabel("%s:" % (self.user_control.label))
		if self.user_control.stylesheet:
			label.setStyleSheet(self.user_control.stylesheet)
		self.layout().addRow(label, self.ctrl)
	
	@property
	def dclass(self):
		
		return self.user_control.dclass
	
	@property
	def descriptor(self):
		
		return self.user_control.descriptor
	
	def set_value(self, value, obj_id = None):
		
		self.ctrl.set_value(value)
		self.ctrl.set_object(obj_id)
	
	def populate_lookup(self):
		
		self.ctrl.populate_lookup()
	
	def get_value(self):
		
		return self.ctrl.get_value()
	
	def get_object(self):
		
		return self.ctrl.obj_id




'''
DMainWindow(QtWidgets.QMainWindow)
	.registry
		.get(var)
		.set(var, value)
		.vars()
		.flush()
'''
from deposit_gui.dgui.dregistry import DRegistry

from PySide2 import (QtWidgets, QtCore, QtGui)

class DMainWindow(QtWidgets.QMainWindow):
	
	APP_NAME = "DView"
	REG_PREFIX = ""
	
	def __init__(self):
		
		self._current_geometry = None
		self._previous_geometry = None
		self._loaded = False
		
		QtWidgets.QMainWindow.__init__(self)
		
		self.registry = DRegistry(self.APP_NAME)
	
	def _load_geometry(self):
		
		if not self.isVisible():
			return
		state = self.registry.get(self.REG_PREFIX + "widow_maximized")
		geometry = self.registry.get(self.REG_PREFIX + "window_geometry")
		restored = False
		if geometry:
			geometry = geometry[1:].strip("'")
			restored = self.restoreGeometry(QtCore.QByteArray().fromPercentEncoding(
				bytearray(geometry, "utf-8"))
			)
		if not restored:
			self.setGeometry(50, 50, 800, 600)
		if state:
			if int(state) == 1:
				self.showMaximized()
			else:
				self.showNormal()
	
	def _get_geometry(self):
		
		return str(self.saveGeometry().toPercentEncoding())
		
	def _save_geometry(self, geometry = None):
		
		geometry = self._get_geometry()
		if self.isVisible():
			if geometry:
				self.registry.set(self.REG_PREFIX + "window_geometry", geometry)
		return geometry
	
	def _save_window_state(self):
		
		if self.isVisible():
			self.registry.set(
				self.REG_PREFIX + "widow_maximized",
				str(int(self.isMaximized()))
			)
	
	# overriden QMainWindow methods
	
	def show(self):
		
		self.setVisible(True)
	
	def setVisible(self, state):
		
		QtWidgets.QMainWindow.setVisible(self, state)
		if state and not self._loaded:
			self._loaded = True
			self._load_geometry()
			self._current_geometry = self._get_geometry()
			self._previous_geometry = self._current_geometry
	
	def changeEvent(self, event):
		
		QtWidgets.QMainWindow.changeEvent(self, event)
		if not self.isVisible():
			return
		if event.type() ==  QtCore.QEvent.WindowStateChange:
			self._save_window_state()
			if self.isMaximized():
				self._save_geometry(self._previous_geometry)
	
	def resizeEvent(self, event):
		
		if self.isVisible():
			self._previous_geometry = self._current_geometry
			self._current_geometry = self._save_geometry()
		QtWidgets.QMainWindow.resizeEvent(self, event)
	
	def moveEvent(self, event):
		
		if self.isVisible():
			self._save_geometry()
		QtWidgets.QMainWindow.moveEvent(self, event)
	
	def closeEvent(self, event):
		
		self.registry.flush()
		QtWidgets.QMainWindow.closeEvent(self, event)




"""
PyGraphviz
==========

A Python wrapper for the graphviz Agraph data structure.

See https://pygraphviz.github.io for complete documentation.
See pygraphviz.AGraph for detailed documentation.
"""

import deposit_gui.dgui
import os
#import sys

path = os.path.normpath(os.path.abspath(os.path.join(os.path.dirname(deposit_gui.dgui.__file__), "graphviz")))
os.add_dll_directory(path)

__version__ = "1.9"

from .agraph import AGraph, Node, Edge, Attribute, ItemAttribute, DotError

__all__ = ["AGraph", "Node", "Edge", "Attribute", "ItemAttribute", "DotError"]

# Per contract with Sphinx-Gallery, this method must be available at top level
from .scraper import _get_sg_image_scraper

#del sys



from deposit_gui.dgui.dcmodel import DCModel

from deposit.datasource import AbstractDatasource

from deposit.utils.fnc_files import (as_url)

from PySide2 import (QtCore)

class CModel(DCModel):
	
	def __init__(self, cmain, store) -> None:
		
		DCModel.__init__(self, cmain, store)
		
		self._update_timer = QtCore.QTimer()
		self._update_timer.setSingleShot(True)
		self._update_timer.timeout.connect(self.on_update_timer)
		self._update_objects = set()
		self._update_classes = set()
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	def on_added(self, objects, classes):
		# elements = [DObject, DClass, ...]
		
		self._update_objects.update(objects)
		self._update_classes.update(classes)
		self._update_timer.start(100)
	
	def on_deleted(self, objects, classes):
		# elements = [obj_id, name, ...]
		
		self._update_objects.update(objects)
		self._update_classes.update(classes)
		self._update_timer.start(100)
	
	def on_changed(self, objects, classes):
		# elements = [DObject, DClass, ...]
		
		self._update_objects.update(objects)
		self._update_classes.update(classes)
		self._update_timer.start(100)
	
	@QtCore.Slot()
	def on_update_timer(self):
		
		if self._update_classes:
			self.cmain.cnavigator.populate_classes()
			self.cmain.cmdiarea.update_class_graphs()
		self.cmain.cmdiarea.update_queries(
			self._update_objects, self._update_classes
		)
		self.cmain.cusertools.on_data_changed()
		self._update_objects.clear()
		self._update_classes.clear()
		self.cmain.cactions.update()
	
	def on_saved(self, datasource):
		
		self.cmain.cview.set_status_message("Saved: %s" % (str(datasource)))
		self.cmain.cactions.update()
	
	def on_loaded(self):
		
		self.update_model_info()
		self.cmain.cmdiarea.close_all()
		self.cmain.cnavigator.populate_classes()
		self.cmain.cnavigator.populate_queries()
		self.cmain.cusertools.populate_tools()
		self.cmain.cusertools.on_data_changed()
		self.cmain.cactions.update()
	
	def on_local_folder_changed(self):
		
		self.update_model_info()
		self.cmain.cactions.update()
	
	def on_queries_changed(self):
		
		self.cmain.cnavigator.populate_queries()
		self.cmain.cactions.update()
	
	def on_user_tools_changed(self):
		
		self.cmain.cusertools.populate_tools()
		self.cmain.cactions.update()
	
	def on_settings_changed(self):
		
		self.cmain.cactions.update()
	
	@QtCore.Slot(str)
	def on_error(self, message):
		
		self.cmain.cview.log_message(message)
		self.cmain.cview.show_warning("Error", message)
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def update_model_info(self):
		
		texts = []
		datasource = self._model.get_datasource()
		texts.append(
			"Data Source: <b>%s (%s)</b>" % (
				datasource.get_name(), str(datasource)
			)
		)
		texts.append("Local Folder: <b>%s</b>" % (str(self._model.get_folder())))
		self.cmain.cmdiarea.set_background_text(
			"".join([("<p>%s</p>" % text) for text in texts])
		)
		self.cmain.cview.set_title(self.get_datasource_name())
	
	def update_recent(self, kwargs):
		
		datasource = kwargs.get("datasource", None)
		if isinstance(datasource, AbstractDatasource):
			kwargs.update(datasource.to_dict())
		
		url = kwargs.get("url", None)
		if not url:
			path = kwargs.get("path", None)
			if path:
				url = as_url(path)
		self.cmain.cview._view.add_recent_connection(
			url = url,
			identifier = kwargs.get("identifier", None),
			connstr = kwargs.get("connstr", None),
		)
	
	def load(self, *args, **kwargs):
		# datasource = Datasource or format
		
		if not self.cmain.check_save():
			return False
		
		return DCModel.load(self, *args, **kwargs)



from deposit_gui.dgui.abstract_subview import AbstractSubview
from deposit_gui.dgui.ddialog import DDialog

from PySide2 import (QtWidgets, QtCore, QtGui)

class DVDialogs(AbstractSubview):
	
	signal_process = QtCore.Signal(str, object)
	signal_cancel = QtCore.Signal(str, object)
	
	def __init__(self, vmain) -> None:
		
		AbstractSubview.__init__(self, vmain)
	
	def open(self, name):
		
		dialog = DDialog(name, self.vmain.get_active_window())
		
		dialog.signal_process.connect(self.on_process)
		dialog.signal_cancel.connect(self.on_cancel)
		
		return dialog
	
	@QtCore.Slot(str, object)
	def on_process(self, name, dialog):
		
		self.signal_process.emit(name, dialog)
	
	@QtCore.Slot(str, object)
	def on_cancel(self, name, dialog):
		
		self.signal_cancel.emit(name, dialog)



from deposit_gui.view.vmdiarea_frames.query_frame_elements.abstract_query_tab import AbstractQueryTab
from deposit_gui.view.vmdiarea_frames.query_frame_elements.abstract_drag_model import AbstractDragModel
from deposit_gui.view.vmdiarea_frames.query_frame_elements.query_item import QueryItem

import deposit_gui
import os

from PySide2 import (QtWidgets, QtCore, QtGui)
from natsort import (natsorted)

class ProxyModel(QtCore.QSortFilterProxyModel):
	
	signal_sorted = QtCore.Signal()
	
	def __init__(self):
		
		super(ProxyModel, self).__init__()
	
	def lessThan(self, source_left, source_right):
		
		values = [("" if val is None else val) for val in [source_left.data(QtCore.Qt.DisplayRole), source_right.data(QtCore.Qt.DisplayRole)]]
		
		return values == natsorted(values)
	
	def sort(self, *args):
		
		QtCore.QSortFilterProxyModel.sort(self, *args)
		
		self.signal_sorted.emit()

class TableModel(AbstractDragModel, QtCore.QAbstractTableModel):
	
	def __init__(self, queryframe):
		
		QtCore.QAbstractTableModel.__init__(self)
		
		self._queryframe = queryframe
		self._query = queryframe._query
		self._proxy_model = ProxyModel()
		self._icons = dict(
			obj = self._queryframe.get_icon("object.svg"),
			geo = self._queryframe.get_icon("geometry.svg"),
			file = self._queryframe.get_icon("file.svg"),
			image = self._queryframe.get_icon("image.svg"),
			remote_file = self._queryframe.get_icon("remote_file.svg"),
			remote_image = self._queryframe.get_icon("remote_image.svg"),
		)
		self.populate_data()
		
		self._proxy_model.setSourceModel(self)
		self._proxy_model.signal_sorted.connect(self._queryframe.on_sorted)
	
	def populate_data(self):
		
		self._query_items = {}
		self._ro_columns = [True] * len(self._query.columns)
		if self._query.main_class not in [None, "*"]:
			self._ro_columns = [(class_name != self._query.main_class) for class_name, _ in self._query.columns]
		
		self._column_names = []
		for class_name, descriptor_name in self._query.columns:
			names = [name for name in [class_name, descriptor_name] if isinstance(name, str)]
			if names:
				self._column_names.append(".".join(names))
			else:
				self._column_names.append("[no class]")
	
	def get_query_item(self, index):
		
		if not index.isValid():
			return QueryItem(index)
		
		row, col = index.row(), index.column()
		
		key = (row, col)
		if key not in self._query_items:
			obj_id, value = self._query[row, col]
			obj_id_row, _ = self._query[row, 0]
			class_name, descriptor_name = self._query.columns[col]
			self._query_items[key] = QueryItem(index, class_name, descriptor_name, obj_id, value, self._icons, self._ro_columns[col], obj_id_row = obj_id_row)
		
		return self._query_items[key]
	
	def rowCount(self, parent):
		
		return len(self._query)
	
	def columnCount(self, parent):

		return len(self._query.columns)
	
	def flags(self, index):
		
		item = self.get_query_item(index)
		
		return AbstractDragModel.flags(self, item)
	
	def headerData(self, section, orientation, role):
		
		if orientation == QtCore.Qt.Horizontal: # column headers
			if role == QtCore.Qt.DisplayRole:
				return self._column_names[section]
			elif role == QtCore.Qt.UserRole:
				return self._query.columns[section]
			return None
		
		if orientation == QtCore.Qt.Vertical: # row headers
			index = self._proxy_model.index(section, 0)
			if role == QtCore.Qt.DisplayRole:
				return str(self.get_query_item(index).obj_id_row)
			elif role == QtCore.Qt.UserRole:
				return self.get_query_item(index).obj_id_row
			
		return None
	
	def data(self, index, role):
		
		if not index.isValid():
			return None
		
		item = self.get_query_item(index)
		
		return item.data(role)
	
	def setData(self, index, value, role = QtCore.Qt.EditRole):
		
		if role == QtCore.Qt.EditRole:
			item = self.get_query_item(index)
			QtCore.QAbstractTableModel.setData(self, index, value, role)
			self._queryframe.on_edited(item, value)
		
		return False
	
	def drag_supported(self, item):
		# item: QueryItem
		
		return True
	
	def drop_supported(self, item):
		# item: QueryItem
		
		return True
	
	def on_drop_url(self, item, urls):
		# item: QueryItem
		
		if urls:
			self._queryframe.on_drop_url(item, urls[0])
		
	def on_drop_text(self, item, text):
		# item: QueryItem
		
		self.setData(item.index, text)
		
	def on_drop_items(self, tgt_item, items):
		# tgt_item: QueryItem
		# items: [QueryItem, ...]
		
		if items:
			self._queryframe.on_edited(tgt_item, items[0].value)

class QueryTabTable(AbstractQueryTab, QtWidgets.QTableView):
	
	def __init__(self, queryframe):
		
		QtWidgets.QTableView.__init__(self)
		
		self._queryframe = queryframe
		self._query = queryframe._query
		self._table_model = TableModel(queryframe)
		
		self.setAcceptDrops(True)
		self.setDragEnabled(True)
		self.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
		self.setDefaultDropAction(QtCore.Qt.CopyAction)
		self.setDropIndicatorShown(True)
		
		self.setSortingEnabled(True)
		self.horizontalHeader().setSortIndicator(0, QtCore.Qt.AscendingOrder)
		
		path = os.path.join(os.path.dirname(deposit_gui.__file__), "res/obj_id.svg")
		path = path.replace("\\", "/")
		self.setStyleSheet('''
			QTableView::item:selected:!active {
				background-color : #81b0d6;
			}
			QTableCornerButton::section {
				background-color: #FFFFFF;
				background-image: url(\"%s\");
				background-position: center;
				background-repeat: no-repeat;
				border: 0px solid lightgrey;
				border-width: 0 1px 0 0;
			}
			QHeaderView::section:vertical {
				width: 50px;
			}
			QHeaderView::section:horizontal {
				height: 30px;
			}
		''' % (path))
		
		self.setIconSize(QtCore.QSize(24, 24))
		
		self.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
		self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
		self.horizontalHeader().setStretchLastSection(True)
		
		self.setModel(self._table_model._proxy_model)
		
		self.horizontalHeader().resizeSections(QtWidgets.QHeaderView.ResizeToContents)
		
		self.activated.connect(self.on_activated)
		self.verticalHeader().sectionDoubleClicked.connect(self.on_row_doubleclicked)
	
	def update_query(self):
		
		self._table_model.beginResetModel()
		self._table_model.populate_data()
		self._table_model.endResetModel()
	
	def apply_filter(self, text):
		
		if text == "":
			self._table_model._proxy_model.setFilterWildcard("")
		else:
			self._table_model._proxy_model.setFilterRegExp(QtCore.QRegExp(".*%s.*" % text, QtCore.Qt.CaseInsensitive))
		self._table_model._proxy_model.setFilterKeyColumn(-1)
	
	def select_object(self, obj_id):
		
		for row in range(self._table_model._proxy_model.rowCount()):
			if self._table_model._proxy_model.index(row, 0).data(QtCore.Qt.UserRole).obj_id_row == obj_id:
				self.selectRow(row)
				return
	
	def set_query_item(self, row, column, value):
		
		self._table_model._proxy_model.setData(self._table_model._proxy_model.index(row, column), value)
	
	def get_row_count(self):
		
		return self._table_model._proxy_model.rowCount()
	
	def get_column_count(self):
		
		return self._table_model._proxy_model.columnCount()
	
	def get_images(self):
		
		images = {}
		n_columns = len(self._query.columns)
		for row in range(len(self._query)):
			for column in range(n_columns):
				value = self._table_model.index(row, column).data(QtCore.Qt.UserRole)
				if isinstance(value, QueryItem):
					value = value.value
					if (value.__class__.__name__ == "DResource") and (value.is_image):
						images[(row, column)] = value
		return images
	
	def get_obj_ids(self):
		# return objects of currently visible rows
		
		obj_ids = set()
		for row in range(self._table_model._proxy_model.rowCount()):
			index = self._table_model._proxy_model.index(row, 0)
			item = self._table_model._proxy_model.mapToSource(index).data(QtCore.Qt.UserRole)
			if isinstance(item, QueryItem):
				obj_id = item.obj_id
				if obj_id is not None:
					obj_ids.add(obj_id)
		return obj_ids
	
	def get_item_order(self):
		# return ordering of items according to current sorting
		#
		# returns item_order[(row, col)] = order
		
		item_order = {}
		n_columns = len(self._query.columns)
		for row in range(self._table_model._proxy_model.rowCount()):
			for column in range(n_columns):
				index = self._table_model._proxy_model.index(row, column)
				index_orig = self._table_model._proxy_model.mapToSource(index)
				item_order[(index_orig.row(), index_orig.column())] = index.row() * n_columns + index.column()
		return item_order
	
	@QtCore.Slot(QtCore.QModelIndex)
	def on_activated(self, index):
		
		item = index.data(QtCore.Qt.UserRole)
		self._queryframe.on_query_activated(item)
	
	@QtCore.Slot(int)
	def on_row_doubleclicked(self, row):
		
		self.on_activated(self._table_model._proxy_model.index(row, 0))
	
	def on_selected(self):
		
		row_items = [index.data(QtCore.Qt.UserRole) for index in self.selectionModel().selectedRows()]
		items = [index.data(QtCore.Qt.UserRole) for index in self.selectionModel().selectedIndexes()]
		items = [item for item in items if item is not None]
		self._queryframe.on_query_selected(items)
		self._queryframe.on_selected_rows(row_items)
	
	def selectionChanged(self, *args, **kwargs):
		
		QtWidgets.QTableView.selectionChanged(self, *args, **kwargs)
		
		self.on_selected()





from deposit_gui.view.vusertools_elements.user_elements.column_break import (ColumnBreak)

from PySide2 import (QtWidgets)

class EditorColumn(QtWidgets.QFrame):
	
	def __init__(self):
		
		QtWidgets.QFrame.__init__(self)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self.setStyleSheet("EditorColumn {border: 1px solid gray;}")
	
	def user_element(self):
		
		return ColumnBreak()



from deposit_gui.dgui.dtool_bar import DToolBar

from PySide2 import (QtWidgets, QtCore, QtGui)

class DMenuBar(DToolBar):
	
	def __init__(self, view):
		
		DToolBar.__init__(self, view)
		
		self.menubar = self.view.menuBar()
		self.menubar.triggered.connect(self.on_triggered)
		
		self.menus = {}  # {name: QMenu, ...}
	
	def set_menu_enabled(self, name, state):
		
		if name not in self.menus:
			return
		self.menus[name].setEnabled(state)
	
	def populate(self, tools):
		# tools = {menu: [(name, caption), None, ...], ...}; if None: add separator
		
		def populate_menu(items, menu):
			
			for name in items:
				if name is None:
					menu.addSeparator()
				elif isinstance(name[1], str):
					name, caption = name
					self.actions[name].append(QtWidgets.QAction(caption, self.view))
					self.actions[name][-1].setData(name)
					menu.addAction(self.actions[name][-1])
				elif isinstance(name[1], list):
					submenu = QtWidgets.QMenu(name[0])
					menu.addMenu(submenu)
					populate_menu(name[1], submenu)
		
		for menu_name in tools:
			self.menus[menu_name] = self.menubar.addMenu(menu_name)
			populate_menu(tools[menu_name], self.menus[menu_name])



from deposit_gui.dgui.abstract_subview import AbstractSubview
from deposit_gui.view.vusertools_elements.user_elements.search_form import (SearchForm)
from deposit_gui.view.vusertools_elements.user_elements.entry_form import (EntryForm)
from deposit_gui.view.vusertools_elements.user_elements.query import (Query)
from deposit_gui.view.vusertools_elements.user_elements import user_controls as UserControls
from deposit_gui.view.vusertools_elements.user_elements import user_groups as UserGroups
from deposit_gui.view.vusertools_elements.user_elements.column_break import (ColumnBreak)
from deposit_gui.view.vusertools_elements.manager import (Manager)
from deposit_gui.view.vusertools_elements.dialog.dialog_search_form import (DialogSearchForm)
from deposit_gui.view.vusertools_elements.dialog.dialog_entry_form import (DialogEntryForm)

from PySide2 import (QtWidgets, QtCore, QtGui)

class VUserTools(AbstractSubview):
	
	signal_search_submit = QtCore.Signal(str)
	#	querystr
	signal_entry_submit = QtCore.Signal(dict, dict, list, set, dict)
	#	values, objects_existing, relations, unique, objects_loaded
	signal_entry_unlink = QtCore.Signal(dict, list, set, list)
	#	objects_existing, relations, unique, obj_ids
	signal_entry_remove = QtCore.Signal(dict, set)
	#	objects_existing, unique
	signal_import_tool = QtCore.Signal()
	signal_export_tool = QtCore.Signal(object)		# user_tool
	signal_add_tool = QtCore.Signal(object)			# user_tool
	signal_update_tool = QtCore.Signal(str, object)	# label, user_tool
	signal_del_tool = QtCore.Signal(list)			# [label, ...]
	
	signal_data_changed = QtCore.Signal()  # internal signal
	signal_tools_changed = QtCore.Signal()  # internal signal
	
	SELECTED_STR = "{selected}"
	
	def __init__(self, vmain, cmodel) -> None:
		
		AbstractSubview.__init__(self, vmain)
		
		self._cmodel = cmodel
		
		self.vmain.addToolBarBreak()
		self._toolbar = self.vmain.addToolBar("User Tools")
		self._manager = None
		self._form_editor = None
		self.entry_form_geometry = None
		
		self._toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
		
		self.action_manage = QtWidgets.QAction("User Tools")
		self.action_manage.setToolTip("Manage User Tools")
		self.action_manage.triggered.connect(self.on_manage)
	
	def populate_tools(self, user_tools):
		
		self._toolbar.clear()
		self._toolbar.addAction(self.action_manage)
		self._toolbar.addSeparator()
		for data in user_tools:
			user_tool = self.user_tool_from_dict(data)
			if user_tool is not None:
				self._toolbar.addAction(user_tool)
		self.signal_tools_changed.emit()
	
	def user_tool_from_dict(self, data):
		
		user_tool = None
		if data["typ"] == "Query":
			return Query(data["label"], data["value"], self)
		else:
			typ = None
			if data["typ"] == "SearchForm":
				typ = SearchForm
			elif data["typ"] == "EntryForm":
				typ = EntryForm
			if typ is not None:
				user_tool = typ(data["label"], self)
				if typ in [SearchForm, EntryForm]:
					for element in data["elements"]:
						if element["typ"] in ["Group", "MultiGroup"]:
							user_tool.elements.append(getattr(UserGroups, element["typ"])(element["stylesheet"], element["label"]))
							for member in element["members"]:
								user_tool.elements[-1].members.append(getattr(UserControls, member["typ"])(member["stylesheet"], member["label"], member["dclass"], member["descriptor"]))
						elif element["typ"] == "ColumnBreak":
							user_tool.elements.append(ColumnBreak())
						elif element["typ"] == "Unique":
							user_tool.elements.append(getattr(UserControls, element["typ"])(element["stylesheet"], element["label"], element["dclass"]))
						else:
							user_tool.elements.append(getattr(UserControls, element["typ"])(element["stylesheet"], element["label"], element["dclass"], element["descriptor"]))
				return user_tool
		return None
	
	def start_manager(self):
		
		self._manager = Manager(self)
		self._manager.signal_export.connect(lambda user_tool: self.signal_export_tool.emit(user_tool))
		self._manager.signal_import.connect(lambda: self.signal_import_tool.emit())
		self._manager.signal_del_tool.connect(lambda labels: self.signal_del_tool.emit(labels))
		self.signal_data_changed.connect(self._manager.on_data_changed)
		self.signal_tools_changed.connect(self._manager.on_tools_changed)
		self._manager.show()
	
	def get_selected_id(self):
		
		objects = self._cmodel.cmain.get_selected_objects()
		if len(objects) == 1:
			return list(objects)[0].id
		return None
	
	def open_query(self, form_tool):
		
		querystr = form_tool.value
		id = str(self.get_selected_id())
		while self.SELECTED_STR in querystr:
			idx1 = querystr.lower().find(self.SELECTED_STR)
			idx2 = idx1 + len(self.SELECTED_STR)
			querystr = querystr[:idx1] + id + querystr[idx2:]
		self.signal_search_submit.emit(querystr)
	
	def open_search_form(self, form_tool):
		
		dialog = DialogSearchForm(self, form_tool)
		dialog.signal_submit.connect(lambda querystr: self.signal_search_submit.emit(querystr))
		self.signal_data_changed.connect(dialog.on_data_changed)
		dialog.update_lookups()
		dialog.show()
	
	def open_entry_form(self, form_tool):
		
		dialog = DialogEntryForm(self, form_tool, self.get_selected_id())
		dialog.signal_submit.connect(
			lambda values, objects_existing, relations, unique, objects_loaded: self.signal_entry_submit.emit(
				values, objects_existing, relations, unique, objects_loaded
			)
		)
		dialog.signal_unlink.connect(
			lambda objects_existing, relations, unique, obj_ids: self.signal_entry_unlink.emit(
				objects_existing, relations, unique, obj_ids
			)
		)
		dialog.signal_remove.connect(lambda objects_existing, unique: self.signal_entry_remove.emit(
				objects_existing, unique
			)
		)
		self.signal_data_changed.connect(dialog.on_data_changed)
		dialog.update_lookups()
		dialog.show()
	
	@QtCore.Slot()
	def on_manage(self):
		
		self.start_manager()
	
	@QtCore.Slot()
	def on_data_changed(self):
		
		self.signal_data_changed.emit()
	
	@QtCore.Slot()
	def on_close(self):
		
		if self._form_editor is not None:
			self._form_editor.close()



from deposit_gui.view.vusertools_elements.user_elements.user_labeled_element import (UserLabeledElement)

class Group(UserLabeledElement):
	
	def __init__(self, stylesheet, label):
		
		self.members = []
		UserLabeledElement.__init__(self, stylesheet, label)
	
	def to_markup(self):
		
		return "<%s style=\"%s\">%s\n" % (self.__class__.__name__, self.stylesheet, self.label) + "\n".join(["\t%s" % (member.to_markup()) for member in self.members]) + "\n</>"
	
	def to_dict(self):
		
		return dict(
			members = [member.to_dict() for member in self.members],
			**UserLabeledElement.to_dict(self),
		)

class MultiGroup(Group):
	
	def __init__(self, stylesheet, label):
		
		Group.__init__(self, stylesheet, label)




from PySide2 import (QtWidgets, QtCore, QtGui)

class DGraphicsView(QtWidgets.QGraphicsView):
	
	def __init__(self):
		
		QtWidgets.QGraphicsView.__init__(self)
		
		self._button_zoom_reset = None
		self._button_frame = QtWidgets.QFrame()
		self._zoomable = True
		self._zoomed = False
		
		scene = QtWidgets.QGraphicsScene(self)
		self.setScene(scene)
		self.setRenderHints(QtGui.QPainter.Antialiasing | QtGui.QPainter.SmoothPixmapTransform)
		self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
		self.setBackgroundBrush(QtCore.Qt.white)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self._button_frame.setLayout(QtWidgets.QHBoxLayout())
		self.layout().addWidget(self._button_frame)
		self.layout().addStretch()
		self.layout().setAlignment(self._button_frame, QtCore.Qt.AlignLeft)
	
	def set_button_zoom_reset(self, caption = "Reset Zoom", icon = None):
		
		if self._button_zoom_reset is not None:
			return
		
		if icon is None:
			self._button_zoom_reset = QtWidgets.QPushButton(caption)
		else:
			self._button_zoom_reset = QtWidgets.QPushButton(icon, "")
			self._button_zoom_reset.setIconSize(QtCore.QSize(24,24))
		self._button_zoom_reset.setToolTip(caption)
		self._button_zoom_reset.hide()
		self._button_frame.layout().addWidget(self._button_zoom_reset)
		self._button_zoom_reset.clicked.connect(self.reset_zoom)
	
	def remove_button_zoom_reset(self):
		
		if self._button_zoom_reset is None:
			return
		self._button_zoom_reset.hide()
		self._button_zoom_reset.deleteLater()
		self._button_zoom_reset = None
	
	def is_zoomable(self):
		
		return self._zoomable
	
	def is_zoomed(self):
		
		return self._zoomed
	
	def set_zoomable(self, state):
		
		self._zoomable = state
	
	def clear(self):
		
		self.scene().clear()
		self.setSceneRect(QtCore.QRectF())
	
	def reset_scene(self):
		
		self.setSceneRect(self.scene().itemsBoundingRect())
	
	@QtCore.Slot()
	def reset_zoom(self):
		
		self._zoomed = False
		rect = self.scene().itemsBoundingRect().marginsAdded(QtCore.QMarginsF(10, 10, 10, 10))
		if rect.isEmpty():
			rect.setRect(-100, -100, 200, 200)
			self.fitInView(rect, QtCore.Qt.KeepAspectRatio)
		else:
			self.fitInView(rect, QtCore.Qt.KeepAspectRatio)
			self.reset_scene()
		if self._button_zoom_reset is not None:
			self._button_zoom_reset.hide()
	
	def scale_view(self, factor, event = None):
		
		if not self._zoomable:
			return
		self._zoomed = True
		if event is not None:
			pos_old = self.mapToScene(event.pos())
		self.scale(factor, factor)
		if event is not None:
			diff = self.mapToScene(event.pos()) - pos_old
			self.translate(diff.x(), diff.y())
		rect = self.scene().itemsBoundingRect().marginsAdded(QtCore.QMarginsF(10, 10, 10, 10))
		rect = rect.united(self.mapToScene(self.frameRect()).boundingRect().toRect())
		self.setSceneRect(rect)
		if self._button_zoom_reset is not None:
			self._button_zoom_reset.setVisible(self.is_zoomed())
	
	def get_scale_factor(self):
		
		trf = self.transform()
		return trf.m11()**0.5
	
	def wheelEvent(self, event):
		
		self.scale_view(2**(event.delta() / 240.0), event)
	
